"""
Initialization of scCODA models

:authors: Johannes Ostner
"""
import numpy as np
import patsy as pt
import importlib

from sccoda.model import dirichlet_models as dm
from sccoda.model import other_models as om
from sccoda.model import dirichlet_time_models as tm


class CompositionalAnalysis:
    """
    Initializer class for compositional models. Please refer to the tutorial for using this class.
    """

    def __new__(cls, data, formula, baseline_index=None, time_column=None):
        """
        Builds count and covariate matrix, returns a CompositionalModel object

        Usage: model = CompositionalAnalysis(data, formula="covariate1 + covariate2", baseline_index="CellTypeA")

        Parameters
        ----------
        data -- anndata object
            anndata object with cell counts as data.X and covariates saved in data.obs
        formula -- string
            R-style formula for building the covariate matrix.
            Categorical covariates are handled automatically, with the covariate value of the first sample being used as the reference category.
            To set a different level as the reference category, use "C(<CovariateName>, Treatment('<ReferenceLevelName>'))"
        baseline_index -- string or int
            Column index that sets the baseline cell type. Can either reference the name of a column or the n-th column (starting at 0)

        Returns
        -------
        model
            A scCODA.models.dirichlet_models.CompositionalModel object
        """

        importlib.reload(dm)

        cell_types = data.var.index.to_list()

        # Get count data
        data_matrix = data.X.astype("float32")

        # Build covariate matrix from R-like formula
        covariate_matrix = pt.dmatrix(formula, data.obs)
        covariate_names = covariate_matrix.design_info.column_names[1:]
        covariate_matrix = covariate_matrix[:, 1:]

        # Invoke instance of the correct model depending on baseline index
        # If baseline index is "simple": Invokes a model for model comparison
        if baseline_index == "simple":
            return om.SimpleModel(covariate_matrix=np.array(covariate_matrix), data_matrix=data_matrix,
                                  cell_types=cell_types, covariate_names=covariate_names, formula=formula)
        # No baseline index
        elif baseline_index is None:
            if time_column is None:
                return dm.NoBaselineModel(covariate_matrix=np.array(covariate_matrix), data_matrix=data_matrix,
                                          cell_types=cell_types, covariate_names=covariate_names, formula=formula)
            else:
                return tm.NoBaselineModelTime(covariate_matrix=np.array(covariate_matrix), data_matrix=data_matrix,
                                              cell_types=cell_types, covariate_names=covariate_names, formula=formula,
                                              time_matrix=data.obs[time_column].to_numpy())
        # Column name as baseline index
        elif baseline_index in cell_types:
            num_index = cell_types.index(baseline_index)
            return dm.BaselineModel(covariate_matrix=np.array(covariate_matrix), data_matrix=data_matrix,
                                    cell_types=cell_types, covariate_names=covariate_names,
                                    baseline_index=num_index, formula=formula)
        # Numeric baseline index
        elif isinstance(baseline_index, int) & (baseline_index < len(cell_types)):
            return dm.BaselineModel(covariate_matrix=np.array(covariate_matrix), data_matrix=data_matrix,
                                    cell_types=cell_types, covariate_names=covariate_names,
                                    baseline_index=baseline_index, formula=formula)
        # None of the above: Throw error
        else:
            raise NameError("Baseline index is not a valid cell type!")



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

import os
import sys
sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'scCODA'
copyright = '2020, Johannes Ostner, Benjamin Schubert'
author = 'Johannes Ostner, Benjamin Schubert'


# -- General configuration ---------------------------------------------------

master_doc = 'index'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['numpydoc',
              'sphinx.ext.autodoc',
              'sphinx.ext.intersphinx',
              'sphinx.ext.doctest',
              'sphinx.ext.coverage',
              'sphinx.ext.mathjax',
              'sphinx.ext.napoleon',
              'sphinx.ext.autosummary']

autodoc_mock_imports = ["tensorflow",
                        "tensorflow_probability"]


def skip(app, what, name, obj, would_skip, options):
    if name == "__init__" or name == "__new__":
        return False
    return would_skip


def setup(app):
    app.connect("autodoc-skip-member", skip)


# Generate the API documentation when building
autosummary_generate = True
autodoc_member_order = 'bysource'
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_use_rtype = True
napoleon_use_param = True
napoleon_custom_sections = [('Params', 'Parameters')]
todo_include_todos = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']



"""
Framework to simulate multiple sets of parameters, and aggregate the results.
The functions to evaluate the results via plots, ... can be found in 'multi_parameter_analysis_functions'

:authors: Johannes Ostner
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import pandas as pd
import pickle as pkl
import datetime
from sklearn.metrics import confusion_matrix

import tensorflow_probability as tfp
import itertools

from sccoda.util import data_generation as gen
from sccoda.model import other_models as om
from sccoda.util import comp_ana as ca

tfd = tfp.distributions
tfb = tfp.bijectors

#%%


class MultiParamSimulation:

    """
    Implements subsequent generation and simulation of datasets with a multitude of parameters,
    such as data dimensions, effect combinations or MCMC chain length.
    """

    def __init__(self, cases=[1], K=[5], n_total=[1000], n_samples=[[5,5]],
                 b_true=[None], w_true=[None], num_results=[10e3], baseline_index=None, formula="x_0"):
        """
        constructor. Simulated Parameters are passed to the constructor as lists, except the type of model, which is fixed for all simulations.
        The simulation is carried out over all possible combinations of specified parameters.
        See the default values for examples

        Parameters
        ----------
        cases -- List
            Number of (binary) covariates
        K -- List
            Number of cell types
        n_total -- List
            number of cells per sample
        n_samples -- List
            number of samples. Each sublist specifies the number of samples for each covariate combination, length 2**cases
        b_true -- List
            Base composition. Each sublist has dimension K.
        w_true -- List
            Effect composition. Each sublist is a nested list that represents a DxK effect matrix
        num_results -- List
            MCMC chain length
        baseline_index -- int
            Index of reference cellltype (None for no baseline)
        formula -- str
            R-style formula used in model specification
        """

        # HMC Settings
        self.n_burnin = int(5e3)  # number of burn-in steps
        self.step_size = 0.01
        self.num_leapfrog_steps = 10

        # All parameter combinations
        self.simulation_params = list(itertools.product(cases, K, n_total, n_samples, b_true, w_true, num_results))

        # Setup result objects
        self.mcmc_results = {}
        self.parameters = pd.DataFrame(
            {'cases': [], 'K': [], 'n_total': [], 'n_samples': [], 'b_true': [], 'w_true': [], 'num_results': []})

        self.baseline_index = baseline_index
        self.formula = formula

        np.set_seed(1234)

    def simulate(self):
        """
        Generation and modeling of single-cell-like data

        Returns
        -------
        """

        i = 0
        # iterate over all parameter combinations

        for c, k, nt, ns, b, w, nr in self.simulation_params:
            # generate data set
            temp_data = gen.generate_case_control(cases=c, K=k, n_total=nt, n_samples=ns, b_true=b, w_true=w)

            # Save parameter set
            s = [c, k, nt, ns, b, w, nr]
            print('Simulating:', s)
            self.parameters.loc[i] = s

            # if baseline model: Simulate with baseline, else: without. The baseline index is always the last one
            ana = ca.CompositionalAnalysis(temp_data, self.formula, baseline_index=self.baseline_index)

            result_temp = ana.sample_hmc(num_results=int(nr), n_burnin=self.n_burnin,
                                         step_size=self.step_size, num_leapfrog_steps=self.num_leapfrog_steps)

            self.mcmc_results[i] = result_temp.summary_prepare()

            i += 1

        return None

    def get_discovery_rates(self):
        """
        Calculates discovery rates and other statistics for each effect, summarized over all entries of beta

        Returns
        -------
        None
            extends self.parameters
        """

        tp = []
        tn = []
        fp = []
        fn = []
        ws = self.parameters.loc[:, "w_true"]

        # For all parameter sets:
        for i in range(len(self.parameters)):

            # Locate modelled slopes
            betas = self.mcmc_results[i][1]["final_parameter"].tolist()
            betas = [0 if b == 0 else 1 for b in betas]

            # Locate ground truth slopes
            wt = [item for sublist in ws[i] for item in sublist]
            wt = [0 if w == 0 else 1 for w in wt]

            # Calculate confusion matrix: beta[d,k]==0 vs. beta[d,k] !=0
            tn_, fp_, fn_, tp_ = confusion_matrix(wt, betas).ravel()

            tp.append(tp_)
            tn.append(tn_)
            fp.append(fp_)
            fn.append(fn_)

        # add results to self.parameters
        self.parameters['tp'] = tp
        self.parameters['tn'] = tn
        self.parameters['fp'] = fp
        self.parameters['fn'] = fn

        return None

    def get_discovery_rates_per_param(self):
        """
        Discovery rates and other statistics for each entry of beta separately. This only works for cases==[1]

        Returns
        -------
        None
            extends self.parameters
        """

        correct = []
        false = []
        ws = self.parameters.loc[:, "w_true"]

        K = len(ws[0][0])

        # For each parameter set:
        for i in range(len(self.parameters)):

            # Locate modelled slopes
            betas = self.mcmc_results[i][1]["final_parameter"].tolist()
            betas = [0 if b == 0 else 1 for b in betas]

            # Locate ground truth slopes
            wt = ws[i][0]
            wt = [0 if w == 0 else 1 for w in wt]

            correct_ = np.zeros(K)
            false_ = np.zeros(K)
            # Count how often each beta[k] is correctly/falsely identified
            for k in range(K):
                if wt[k] == betas[k]:
                    correct_[k] += 1
                else:
                    false_[k] += 1

            correct.append(correct_)
            false.append(false_)

        # Add results to self.paramerters
        for i in range(K):
            self.parameters['correct_'+str(i)] = [correct[n][i] for n in range(len(correct))]
            self.parameters['false_'+str(i)] = [false[n][i] for n in range(len(false))]

        return None

    def plot_discovery_rates(self, dim_1='w_true', dim_2='n_samples'):
        """
        plots TPR and TNR for two parameter series specified in the constructor (e.g. w_true vs. n_samples)

        Parameters
        ----------
        dim_1 -- str
            parameter name on x-axis
        dim_2 -- str
            parameter name on y-axis

        Returns
        -------

        """

        plot_data = self.parameters[[dim_1, dim_2, "tpr_mcmc", "tnr_mcmc"]]
        plot_data[[dim_1, dim_2]] = plot_data[[dim_1, dim_2]].astype(str)

        fig, ax = plt.subplots(1, 2)
        sns.heatmap(plot_data.pivot(dim_1, dim_2, 'tpr_mcmc'), ax=ax[0]).set_title("MCMC TPR")
        sns.heatmap(plot_data.pivot(dim_1, dim_2, 'tnr_mcmc'), ax=ax[1]).set_title("MCMC TNR")
        plt.show()

    def save(self, path='''/Users/Johannes/Documents/Uni/Master's Thesis/simulation_results/tests/''',
             filename=str(datetime.datetime.now())):
        """
        saves results to a pickle file

        Parameters
        ----------
        path -- str
            directory
        filename -- str
            name of new file

        Returns
        -------

        """

        with open(path + filename + '.pkl', 'wb') as f:
            pkl.dump(self, f, protocol=4)


class MultiParamSimulationMultiModel:

    """
    Implements subsequent simulation of parameter sets with multiple models.
    This class provides a framework for running an instance of multi_param_simulation with more than one model.
    Each generated dataset is evaluated by each model
    """

    def __init__(self, cases=[1], K=[5], n_total=[1000], n_samples=[[5,5]],
                 b_true=[None], w_true=[None], num_results=[10e3], models=["NoBaseline"]):
        """

        Parameters
        ----------
        cases -- list
            Number of (binary) covariates
        K -- list
            Number of cell types
        n_total -- list
            number of cells per sample
        n_samples -- list
            number of samples. Each sublist specifies the number of samples for each covariate combination, length 2**cases
        b_true -- list
            Base composition. Each sublist has dimension K.
        w_true -- list
            Effect composition. Each sublist is a nested list that represents a DxK effect matrix
        num_results -- list
            MCMC chain length
        models -- list
            List of models to evaluate
        formula -- str
            R-like model formula
        """

        # HMC Settings
        self.n_burnin = int(5e3)  # number of burn-in steps
        self.step_size = 0.01
        self.num_leapfrog_steps = 10

        # All parameter combinations
        self.l = list(itertools.product(cases, K, n_total, n_samples, b_true, w_true, num_results))
        self.formula = "x_0"

        # Setup result objects
        self.results = {}
        self.parameters = pd.DataFrame(
            {'cases': [], 'K': [], 'n_total': [], 'n_samples': [], 'b_true': [], 'w_true': [], 'num_results': []})

        self.models = models
        self.data = {}

    def simulate(self):
        """
        Generation and modeling of single-cell-like data

        Returns
        -------
        None
            Fills up self.mcmc_results
        """

        for j in range(len(self.models)):
            self.results[j] = {}

        i = 0

        # For each parameter combination:
        for c, k, nt, ns, b, w, nr in self.l:
            # Generate dataset
            temp_data = gen.generate_case_control(cases=c, K=k, n_total=nt, n_samples=ns,
                                                  b_true=b, w_true=w, sigma=np.identity(k) * 0.01)

            self.data[i] = temp_data

            x_temp = temp_data.obs.values
            y_temp = temp_data.X

            # Write parameter combination
            s = [c, k, nt, ns, b, w, nr]
            print('Simulating:', s)
            self.parameters.loc[i] = s

            j = 0

            # For each model:
            for model in self.models:
                # If Poisson model: Simulate, eval Poisson
                if model == "Poisson":
                    print("Model: Poisson")
                    # Catch edge case of perfect separation
                    if ns == [1, 1]:
                        self.results[j][i] = (1, 4, 0, 0)
                    else:
                        model_temp = om.PoissonModel(covariate_matrix=x_temp, data_matrix=y_temp)
                        model_temp.fit_model()
                        tp, tn, fp, fn = model_temp.eval_model()
                        self.results[j][i] = (tp, tn, fp, fn)

                # If simple model: Simulate, set "final_parameter" to 0 if 95% confint includes 0
                elif model == "Simple":
                    print("Model: Simple")
                    ana = ca.CompositionalAnalysis(temp_data, self.formula, baseline_index="simple")
                    result_temp = ana.sample_hmc(num_results=int(nr), n_burnin=self.n_burnin,
                                                 step_size=self.step_size, num_leapfrog_steps=self.num_leapfrog_steps)
                    alphas_df, betas_df = result_temp.summary_prepare(credible_interval=0.95)

                    betas_df.loc[:, "final_parameter"] = np.where((betas_df.loc[:, "hpd_2.5%"] < 0) &
                                                                  (betas_df.loc[:, "hpd_97.5%"] > 0),
                                                                  0,
                                                                  betas_df.loc[:, "final_parameter"])

                    self.results[j][i] = (alphas_df, betas_df)

                # if baseline model: Simulate with baseline, else: without. The baseline index is always the last one
                elif model == "Baseline":
                    print("Model: Baseline")
                    ana = ca.CompositionalAnalysis(temp_data, self.formula, baseline_index=k-1)
                    result_temp = ana.sample_hmc(num_results=int(nr), n_burnin=self.n_burnin,
                                                 step_size=self.step_size, num_leapfrog_steps=self.num_leapfrog_steps)
                    self.results[j][i] = result_temp.summary_prepare()

                elif model == "NoBaseline":
                    print("Model: No Baseline")
                    ana = ca.CompositionalAnalysis(temp_data, self.formula, baseline_index=None)
                    result_temp = ana.sample_hmc(num_results=int(nr), n_burnin=self.n_burnin,
                                                 step_size=self.step_size, num_leapfrog_steps=self.num_leapfrog_steps)
                    self.results[j][i] = result_temp.summary_prepare()

                # If SCDC model: Export data, run R script
                elif model == "SCDC":
                    print("model: SCDC")
                    model = om.scdney_model(data=temp_data, ns=ns)
                    r = model.analyze()
                    self.results[j][i] = r

                else:
                    print("Not a valid model specified")

                # HMC sampling, save results

                j += 1

            i += 1

        return None

    def get_discovery_rates(self):
        """
        Calculates discovery rates and other statistics for each effect, summarized over all entries of beta

        Returns
        -------
        None
            extends self.parameters
        """

        def get_discovery_rate(res):
            tp = []
            tn = []
            fp = []
            fn = []
            ws = self.parameters.loc[:, "w_true"]

            # For all parameter sets:
            for i in range(len(self.parameters)):
                # Locate modelled slopes
                betas = res[i][1]["final_parameter"].tolist()
                betas = [0 if b == 0 else 1 for b in betas]

                # Locate ground truth slopes
                wt = [item for sublist in ws[i] for item in sublist]
                wt = [0 if w == 0 else 1 for w in wt]

                # Calculate confusion matrix: beta[d,k]==0 vs. beta[d,k] !=0
                tn_, fp_, fn_, tp_ = confusion_matrix(wt, betas).ravel()

                tp.append(tp_)
                tn.append(tn_)
                fp.append(fp_)
                fn.append(fn_)

            return tp, tn, fp, fn

        # add results to self.parameters
        for j in range(len(self.models)):
            if self.models[j] == "Poisson" or self.models[j] == "SCDC":
                tp = []
                tn = []
                fp = []
                fn = []

                if self.models[j] == "SCDC":
                    for res in self.results[j].values():
                        res_ = res[1]
                        tp.append(res_[0])
                        tn.append(res_[1])
                        fp.append(res_[2])
                        fn.append(res_[3])
                else:
                    for res in self.results[j].values():
                        tp.append(res[0])
                        tn.append(res[1])
                        fp.append(res[2])
                        fn.append(res[3])

                self.parameters['tp_' + str(j)] = tp
                self.parameters['tn_' + str(j)] = tn
                self.parameters['fp_' + str(j)] = fp
                self.parameters['fn_' + str(j)] = fn
            else:
                rates = get_discovery_rate(self.results[j])
                self.parameters['tp_'+str(j)] = rates[0]
                self.parameters['tn_'+str(j)] = rates[1]
                self.parameters['fp_'+str(j)] = rates[2]
                self.parameters['fn_'+str(j)] = rates[3]

        return None

    def save(self, path='''/Users/Johannes/Documents/Uni/Master's Thesis/simulation_results/tests/''',
             filename=str(datetime.datetime.now())):
        """
        saves results to a pickle file

        Parameters
        ----------
        path -- str
            directory
        filename -- str
            name of new file

        Returns
        -------

        """
        with open(path + filename + '.pkl', 'wb') as f:
            pkl.dump(self, f, protocol=4)



"""
Dirichlet-multinomial models for statistical analysis of compositional changes

For further reference, see:
Johannes Ostner: Development of a statistical framework for compositional analysis of single-cell data

:authors: Johannes Ostner
"""
import numpy as np
import time

import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow_probability.python.experimental import edward2 as ed

from sccoda.util import result_classes as res

tfd = tfp.distributions
tfb = tfp.bijectors


class CompositionalModel:
    """
    Implements class framework for compositional data models
    """

    def __init__(self, covariate_matrix, data_matrix, cell_types, covariate_names, formula, time_matrix=None, *args, **kwargs):
        """
        Generalized Constructor of model class

        Parameters
        ----------
        covariate_matrix -- numpy array [NxD]
            covariate matrix
        data_matrix -- numpy array [NxK]
            cell count matrix
        cell_types -- list
            Cell type names
        covariate_names -- List
            Covariate names
        """

        dtype = tf.float32
        self.x = tf.convert_to_tensor(covariate_matrix, dtype)

        # Add pseudocount if needed.
        if np.count_nonzero(data_matrix) != np.size(data_matrix):
            print("Zero counts encountered in data! Added a pseudocount of 0.5.")
            data_matrix += 0.5
        self.y = tf.convert_to_tensor(data_matrix, dtype)

        sample_counts = np.sum(data_matrix, axis=1)
        self.n_total = tf.cast(sample_counts, dtype)
        self.cell_types = cell_types
        self.covariate_names = covariate_names
        self.formula = formula

        if time_matrix is not None:
            self.time_matrix = tf.convert_to_tensor(time_matrix, dtype)

        # Get dimensions of data
        self.N, self.D = self.x.shape
        self.K = self.y.shape[1]

        # Check input data
        if self.N != self.y.shape[0]:
            raise ValueError("Wrong input dimensions X[{},:] != y[{},:]".format(self.x.shape[0], self.y.shape[0]))
        if self.N != len(self.n_total):
            raise ValueError("Wrong input dimensions X[{},:] != n_total[{}]".format(self.x.shape[0], len(self.n_total)))

    def sampling(self, num_results, n_burnin, kernel, init_state, trace_fn):
        """
        MCMC sampling process

        Parameters
        ----------
        num_results -- int
            MCMC chain length (default 20000)
        n_burnin -- int
            Number of burnin iterations (default 5000)
        kernel --
            tensorflow MCMC kernel object
        init_state -- dict
            Starting parameters

        Returns
        -------
        states -- list
            States of MCMC chain
        kernel_results -- list
            sampling meta-information
        duration -- float
            Duration of MCMC sampling process
        """

        # HMC sampling function
        @tf.function
        def sample_mcmc(num_results_, n_burnin_, kernel_, current_state_, trace_fn):

            return tfp.mcmc.sample_chain(
                num_results=num_results_,
                num_burnin_steps=n_burnin_,
                kernel=kernel_,
                current_state=current_state_,
                trace_fn=trace_fn
            )

        # The actual sampling process
        start = time.time()
        states, kernel_results = sample_mcmc(num_results, n_burnin, kernel, init_state, trace_fn)
        duration = time.time() - start
        print("MCMC sampling finished. ({:.3f} sec)".format(duration))

        return states, kernel_results, duration

    def get_chains_after_burnin(self, samples, kernel_results, n_burnin, is_nuts=False):
        """
        Application of burnin after sampling

        Parameters
        ----------
        samples -- list
            all kernel states
        kernel_results  -- list
            Kernel meta-information
        n_burnin -- int
            number of burnin iterations

        Returns
        -------
        states_burnin -- list
            Kernel states without burnin samples
        p_accept -- float
            acceptance rate of MCMC process
        """
        # Samples after burn-in
        states_burnin = []
        stats = {}

        for s in samples:
            states_burnin.append(s[n_burnin:].numpy())

        for k, v in kernel_results.items():
            stats[k] = v[n_burnin:].numpy()

        if is_nuts:
            p_accept = np.mean(np.exp(kernel_results["log_accept_ratio"].numpy()))
        else:
            acceptances = kernel_results["is_accepted"].numpy()

            # Calculate acceptance rate
            p_accept = sum(acceptances) / acceptances.shape[0]
        print('Acceptance rate: %0.1f%%' % (100 * p_accept))

        return states_burnin, stats, p_accept

    def sample_hmc(self, num_results=int(20e3), n_burnin=int(5e3), num_leapfrog_steps=10, step_size=0.01, num_adapt_steps=None):
        """
        HMC sampling

        Parameters
        ----------
        num_results -- int
            MCMC chain length (default 20000)
        n_burnin -- int
            Number of burnin iterations (default 5000)
        num_leapfrog_steps --  int
            HMC leapfrog steps (default 10)
        step_size -- float
            Initial step size (default 0.01)

        Returns
        -------
        result -- scCODA.util.result_data.CAResult object
            Compositional analysis result
        """

        # bijectors (not in use atm, therefore identity)
        constraining_bijectors = [tfb.Identity() for x in range(len(self.params))]

        # HMC transition kernel
        hmc_kernel = tfp.mcmc.HamiltonianMonteCarlo(
            target_log_prob_fn=self.target_log_prob_fn,
            step_size=step_size,
            num_leapfrog_steps=num_leapfrog_steps)
        hmc_kernel = tfp.mcmc.TransformedTransitionKernel(
            inner_kernel=hmc_kernel, bijector=constraining_bijectors)

        # Set default value for adaptation steps
        if num_adapt_steps is None:
            num_adapt_steps = int(0.8 * n_burnin)

        # Add step size adaptation
        hmc_kernel = tfp.mcmc.SimpleStepSizeAdaptation(
            inner_kernel=hmc_kernel, num_adaptation_steps=num_adapt_steps, target_accept_prob=0.8)

        # tracing function
        def trace_fn(_, pkr):
            return {
                'target_log_prob': pkr.inner_results.inner_results.accepted_results.target_log_prob,
                'diverging': (pkr.inner_results.inner_results.log_accept_ratio < -1000.),
                'is_accepted': pkr.inner_results.inner_results.is_accepted,
                'step_size': pkr.inner_results.inner_results.accepted_results.step_size,
            }

        # HMC sampling
        states, kernel_results, duration = self.sampling(num_results, n_burnin, hmc_kernel, self.params, trace_fn)

        # apply burnin
        states_burnin, sample_stats, acc_rate = self.get_chains_after_burnin(states, kernel_results, n_burnin)

        # Calculate posterior predictive
        y_hat = self.get_y_hat(states_burnin, num_results, n_burnin)

        params = dict(zip(self.param_names, states_burnin))

        # Result object generation setup
        # Get names of cell types that are not the baseline
        if self.baseline_index is not None:
            cell_types_nb = self.cell_types[:self.baseline_index] + self.cell_types[self.baseline_index+1:]
        else:
            cell_types_nb = self.cell_types

        posterior = {var_name: [var] for var_name, var in params.items() if
                     "prediction" not in var_name}

        if "prediction" in self.param_names:
            posterior_predictive = {"prediction": [params["prediction"]]}
        else:
            posterior_predictive = {}

        observed_data = {"y": self.y}
        dims = {"alpha": ["cell_type"],
                "mu_b": ["1"],
                "sigma_b": ["1"],
                "b_offset": ["covariate", "cell_type_nb"],
                "ind_raw": ["covariate", "cell_type_nb"],
                "ind": ["covariate", "cell_type_nb"],
                "b_raw": ["covariate", "cell_type_nb"],
                "beta": ["covariate", "cell_type"],
                "concentration": ["sample", "cell_type"],
                "prediction": ["sample", "cell_type"]
                }
        coords = {"cell_type": self.cell_types,
                  "cell_type_nb": cell_types_nb,
                  "covariate": self.covariate_names,
                  "sample": range(self.y.shape[0])
                  }

        sampling_stats = {"chain_length": num_results, "n_burnin": n_burnin,
                        "acc_rate": acc_rate, "duration": duration, "y_hat": y_hat}

        model_specs = {"baseline": self.baseline_index, "formula": self.formula}

        return res.CAResultConverter(posterior=posterior,
                                     posterior_predictive=posterior_predictive,
                                     observed_data=observed_data,
                                     dims=dims,
                                     sample_stats=sample_stats,
                                     coords=coords).to_result_data(sampling_stats=sampling_stats,
                                                                   model_specs=model_specs)

    def sample_hmc_da(self, num_results=int(20e3), n_burnin=int(5e3), num_leapfrog_steps=10, step_size=0.01, num_adapt_steps=None):
        """
        HMC sampling

        Parameters
        ----------
        num_results -- int
            MCMC chain length (default 20000)
        n_burnin -- int
            Number of burnin iterations (default 5000)
        num_leapfrog_steps --  int
            HMC leapfrog steps (default 10)
        step_size -- float
            Initial step size (default 0.01)

        Returns
        -------
        result -- scCODA.util.result_data.CAResult object
            Compositional analysis result
        """

        # (not in use atm)
        constraining_bijectors = [tfb.Identity() for x in range(len(self.params))]

        # HMC transition kernel
        hmc_kernel = tfp.mcmc.HamiltonianMonteCarlo(
            target_log_prob_fn=self.target_log_prob_fn,
            step_size=step_size,
            num_leapfrog_steps=num_leapfrog_steps)
        hmc_kernel = tfp.mcmc.TransformedTransitionKernel(
            inner_kernel=hmc_kernel, bijector=constraining_bijectors)

        # Set default value for adaptation steps
        if num_adapt_steps is None:
            num_adapt_steps = int(0.8*n_burnin)

        # Add step size adaptation
        hmc_kernel = tfp.mcmc.DualAveragingStepSizeAdaptation(
            inner_kernel=hmc_kernel, num_adaptation_steps=num_adapt_steps, target_accept_prob=0.8, decay_rate=0.5)

        # tracing function
        def trace_fn(_, pkr):
            return {
                'target_log_prob': pkr.inner_results.inner_results.accepted_results.target_log_prob,
                'diverging': (pkr.inner_results.inner_results.log_accept_ratio < -1000.),
                "log_acc_ratio": pkr.inner_results.inner_results.log_accept_ratio,
                'is_accepted': pkr.inner_results.inner_results.is_accepted,
                'step_size': tf.exp(pkr.log_averaging_step[0]),
            }

        # HMC sampling
        states, kernel_results, duration = self.sampling(num_results, n_burnin, hmc_kernel, self.params, trace_fn)
        states_burnin, sample_stats, acc_rate = self.get_chains_after_burnin(states, kernel_results, n_burnin)

        y_hat = self.get_y_hat(states_burnin, num_results, n_burnin)

        params = dict(zip(self.param_names, states_burnin))

        # Result object generation setup
        if self.baseline_index is not None:
            cell_types_nb = self.cell_types[:self.baseline_index] + self.cell_types[self.baseline_index+1:]
        else:
            cell_types_nb = self.cell_types

        posterior = {var_name: [var] for var_name, var in params.items() if
                     "prediction" not in var_name}

        if "prediction" in self.param_names:
            posterior_predictive = {"prediction": [params["prediction"]]}
        else:
            posterior_predictive = {}

        observed_data = {"y": self.y}
        dims = {"alpha": ["cell_type"],
                "mu_b": ["1"],
                "sigma_b": ["1"],
                "b_offset": ["covariate", "cell_type_nb"],
                "ind_raw": ["covariate", "cell_type_nb"],
                "ind": ["covariate", "cell_type_nb"],
                "b_raw": ["covariate", "cell_type_nb"],
                "beta": ["covariate", "cell_type"],
                "concentration": ["sample", "cell_type"],
                "prediction": ["sample", "cell_type"]
                }
        coords = {"cell_type": self.cell_types,
                  "cell_type_nb": cell_types_nb,
                  "covariate": self.covariate_names,
                  "sample": range(self.y.shape[0])
                  }

        # build dicionary with sampling statistics
        sampling_stats = {"chain_length": num_results, "n_burnin": n_burnin,
                        "acc_rate": acc_rate, "duration": duration, "y_hat": y_hat}

        model_specs = {"baseline": self.baseline_index, "formula": self.formula}

        return res.CAResultConverter(posterior=posterior,
                                     posterior_predictive=posterior_predictive,
                                     observed_data=observed_data,
                                     dims=dims,
                                     sample_stats=sample_stats,
                                     coords=coords).to_result_data(sampling_stats=sampling_stats,
                                                                   model_specs=model_specs)

    def sample_nuts(self, num_results=int(10e3), n_burnin=int(5e3), max_tree_depth=10, step_size=0.01, num_adapt_steps=None):
        """
        NUTS sampling - WIP, DO NOT USE!!!

        Parameters
        ----------
        num_results -- int
            MCMC chain length (default 20000)
        n_burnin -- int
            Number of burnin iterations (default 5000)
        max_tre_depth --  int
            Maximum tree depth (default 10)
        step_size -- float
            Initial step size (default 0.01)

        Returns
        -------
        error
            NotImplementedError
        """

        # bijectors (not in use atm, therefore identity)
        constraining_bijectors = [tfb.Identity() for x in range(len(self.params))]

        # NUTS transition kernel
        nuts_kernel = tfp.mcmc.NoUTurnSampler(
            target_log_prob_fn=self.target_log_prob_fn,
            step_size=step_size,
            max_tree_depth=max_tree_depth)
        nuts_kernel = tfp.mcmc.TransformedTransitionKernel(
            inner_kernel=nuts_kernel,
            bijector=constraining_bijectors
        )

        # Set default value for adaptation steps
        if num_adapt_steps is None:
            num_adapt_steps = int(0.8 * n_burnin)

        # Step size adaptation
        nuts_kernel = tfp.mcmc.DualAveragingStepSizeAdaptation(
            inner_kernel=nuts_kernel, num_adaptation_steps=num_adapt_steps, target_accept_prob=0.8,
            step_size_setter_fn=lambda pkr, new_step_size: pkr._replace(
                inner_results=pkr.inner_results._replace(step_size=new_step_size)
            ),
            step_size_getter_fn=lambda pkr: pkr.inner_results.step_size,
            log_accept_prob_getter_fn=lambda pkr: pkr.inner_results.log_accept_ratio,
        )

        # trace function
        def trace_fn(_, pkr):
            return {
                "target_log_prob": pkr.inner_results.inner_results.target_log_prob,
                "leapfrogs_taken": pkr.inner_results.inner_results.leapfrogs_taken,
                "diverging": pkr.inner_results.inner_results.has_divergence,
                "energy": pkr.inner_results.inner_results.energy,
                "log_accept_ratio": pkr.inner_results.inner_results.log_accept_ratio,
                'step_size': pkr.inner_results.inner_results.step_size[0],
                "reach_max_depth": pkr.inner_results.inner_results.reach_max_depth,
                "is_accepted": pkr.inner_results.inner_results.is_accepted,
            }

        # HMC sampling
        states, kernel_results, duration = self.sampling(num_results, n_burnin, nuts_kernel, self.params, trace_fn)
        states_burnin, sample_stats, acc_rate = self.get_chains_after_burnin(states, kernel_results, n_burnin, is_nuts=True)

        y_hat = self.get_y_hat(states_burnin, num_results, n_burnin)

        params = dict(zip(self.param_names, states_burnin))

        # Result object generation setup
        # Get names of cell types that are not the baseline
        if self.baseline_index is not None:
            cell_types_nb = self.cell_types[:self.baseline_index] + self.cell_types[self.baseline_index + 1:]
        else:
            cell_types_nb = self.cell_types

        posterior = {var_name: [var] for var_name, var in params.items() if
                     "prediction" not in var_name}

        if "prediction" in self.param_names:
            posterior_predictive = {"prediction": [params["prediction"]]}
        else:
            posterior_predictive = {}

        observed_data = {"y": self.y}
        dims = {"alpha": ["cell_type"],
                "mu_b": ["1"],
                "sigma_b": ["1"],
                "b_offset": ["covariate", "cell_type_nb"],
                "ind_raw": ["covariate", "cell_type_nb"],
                "ind": ["covariate", "cell_type_nb"],
                "b_raw": ["covariate", "cell_type_nb"],
                "beta": ["covariate", "cell_type"],
                "concentration": ["sample", "cell_type"],
                "prediction": ["sample", "cell_type"]
                }
        coords = {"cell_type": self.cell_types,
                  "cell_type_nb": cell_types_nb,
                  "covariate": self.covariate_names,
                  "sample": range(self.y.shape[0])
                  }

        sampling_stats = {"chain_length": num_results, "n_burnin": n_burnin,
                          "acc_rate": acc_rate, "duration": duration, "y_hat": y_hat}

        model_specs = {"baseline": self.baseline_index, "formula": self.formula}

        return res.CAResultConverter(posterior=posterior,
                                     posterior_predictive=posterior_predictive,
                                     observed_data=observed_data,
                                     dims=dims,
                                     sample_stats=sample_stats,
                                     coords=coords).to_result_data(sampling_stats=sampling_stats,
                                                                   model_specs=model_specs)


class NoBaselineModel(CompositionalModel):

    """"
    implements statistical model for compositional differential change analysis without specification of a baseline cell type
    """

    def __init__(self, *args, **kwargs):
        """
        Constructor of model class

        Parameters
        ----------
        args -- arguments passed to top-level class
        kwargs -- arguments passed to top-level class
        """
        super(self.__class__, self).__init__(*args, **kwargs)

        self.baseline_index = None
        dtype = tf.float32

        # All parameters that are returned for analysis
        self.param_names = ["alpha", "mu_b", "sigma_b", "b_offset", "ind_raw",
                            "ind", "b_raw", "beta", "concentration", "prediction"]

        def define_model(x, n_total, K):
            """
            Model definition in Edward2

            Parameters
            ----------
            x -- numpy array [NxD]
                covariate matrix
            n_total -- numpy array [N]
                number of cells per sample
            K -- int
                Number of cell types
            """

            N, D = x.shape
            dtype = tf.float32

            # normal prior on bias
            alpha = ed.Normal(loc=tf.zeros([K]), scale=tf.ones([K])*5, name="alpha")

            # Noncentered parametrization for raw slopes (before spike-and-slab)
            mu_b = ed.Normal(loc=tf.zeros(1, dtype=dtype), scale=tf.ones(1, dtype=dtype), name="mu_b")
            sigma_b = ed.HalfCauchy(tf.zeros(1, dtype=dtype), tf.ones(1, dtype=dtype), name="sigma_b")
            b_offset = ed.Normal(loc=tf.zeros([D, K], dtype=dtype), scale=tf.ones([D, K], dtype=dtype), name="b_offset")

            b_raw = mu_b + sigma_b * b_offset

            # Spike-and-slab priors
            sigma_ind_raw = ed.Normal(
                loc=tf.zeros(shape=[D, K], dtype=dtype),
                scale=tf.ones(shape=[D, K], dtype=dtype),
                name='sigma_ind_raw')
            ind_t = sigma_ind_raw*50
            ind = tf.exp(ind_t) / (1 + tf.exp(ind_t))

            # Calculate betas
            beta = ind * b_raw

            # Concentration vector from intercepts, slopes
            concentration_ = tf.exp(alpha + tf.matmul(x, beta))

            # Cell count prediction via DirMult
            predictions = ed.DirichletMultinomial(n_total, concentration=concentration_, name="predictions")
            return predictions

        # Joint posterior distribution
        self.log_joint = ed.make_log_joint_fn(define_model)

        # Function to compute log posterior probability
        self.target_log_prob_fn = lambda alpha_, mu_b_, sigma_b_, b_offset_, sigma_ind_raw_:\
            self.log_joint(x=self.x,
                           n_total=self.n_total,
                           K=self.K,
                           predictions=self.y,
                           alpha=alpha_,
                           mu_b=mu_b_,
                           sigma_b=sigma_b_,
                           b_offset=b_offset_,
                           sigma_ind_raw=sigma_ind_raw_,
                           )

        alpha_size = [self.K]
        beta_size = [self.D, self.K]

        # MCMC starting values
        self.params = [tf.random.normal(alpha_size, 0, 1, name='init_alpha'),
                       tf.zeros(1, name="init_mu_b", dtype=dtype),
                       tf.ones(1, name="init_sigma_b", dtype=dtype),
                       tf.random.normal(beta_size, 0, 1, name='init_b_offset'),
                       tf.zeros(beta_size, name='init_sigma_ind_raw', dtype=dtype),
                       ]

    # Calculate predicted cell counts (for analysis purposes)
    def get_y_hat(self, states_burnin, num_results, n_burnin):
        """
        Calculate predicted cell counts (for analysis purposes) and add intermediate parameters to MCMC results

        Parameters
        ----------
        states_burnin -- List
            MCMC chain without burnin samples
        num_results -- int
            Chain length (with burnin)
        n_burnin -- int
            Number of burnin samples

        Returns
        -------
        y_mean
            predicted cell counts
        """

        chain_size_y = [num_results - n_burnin, self.N, self.K]

        alphas = states_burnin[0]
        alphas_final = alphas.mean(axis=0)

        ind_raw = states_burnin[4] * 50
        mu_b = states_burnin[1]
        sigma_b = states_burnin[2]
        b_offset = states_burnin[3]

        ind_ = np.exp(ind_raw) / (1 + np.exp(ind_raw))

        b_raw_ = mu_b.reshape((num_results - n_burnin, 1, 1)) + np.einsum("...jk, ...j->...jk", b_offset, sigma_b)

        beta_ = np.einsum("..., ...", ind_, b_raw_)

        conc_ = np.exp(np.einsum("jk, ...kl->...jl", self.x, beta_)
                       + alphas.reshape((num_results - n_burnin, 1, self.K)))

        predictions_ = np.zeros(chain_size_y)
        for i in range(num_results-n_burnin):
            pred = tfd.DirichletMultinomial(self.n_total, conc_[i, :, :]).mean().numpy()
            predictions_[i, :, :] = pred

        betas_final = beta_.mean(axis=0)
        states_burnin.append(ind_)
        states_burnin.append(b_raw_)
        states_burnin.append(beta_)
        states_burnin.append(conc_)
        states_burnin.append(predictions_)

        concentration = np.exp(np.matmul(self.x, betas_final) + alphas_final).astype(np.float32)
        y_mean = concentration / np.sum(concentration, axis=1, keepdims=True) * self.n_total.numpy()[:, np.newaxis]
        return y_mean


class BaselineModel(CompositionalModel):
    """
    implements statistical model for compositional differential change analysis with specification of a baseline cell type
    """

    def __init__(self, baseline_index, *args, **kwargs):

        """
        Constructor of model class

        Parameters
        ----------
        baseline_index -- string or int
            Index of reference cell type (column in count data matrix)
        args -- arguments passed to top-level class
        kwargs -- arguments passed to top-level class
        """
        super(self.__class__, self).__init__(*args, **kwargs)

        self.baseline_index = baseline_index
        dtype = tf.float32

        # All parameters that are returned for analysis
        self.param_names = ["alpha", "mu_b", "sigma_b", "b_offset", "ind_raw",
                            "ind", "b_raw", "beta", "concentration", "prediction"]

        def define_model(x, n_total, K):
            """
            Model definition in Edward2

            Parameters
            ----------
            x -- numpy array [NxD]
                covariate matrix
            n_total -- numpy array [N]
                number of cells per sample
            K -- int
                Number of cell types
            """
            dtype = tf.float32
            N, D = x.shape

            # normal prior on bias
            alpha = ed.Normal(loc=tf.zeros([K]), scale=tf.ones([K]) * 5, name="alpha")

            # Noncentered parametrization for raw slopes of all cell types except baseline type (before spike-and-slab)
            mu_b = ed.Normal(loc=tf.zeros(1, dtype=dtype), scale=tf.ones(1, dtype=dtype), name="mu_b")
            sigma_b = ed.HalfCauchy(tf.zeros(1, dtype=dtype), tf.ones(1, dtype=dtype), name="sigma_b")
            b_offset = ed.Normal(loc=tf.zeros([D, K-1], dtype=dtype), scale=tf.ones([D, K-1], dtype=dtype),
                                 name="b_offset")

            b_raw = mu_b + sigma_b * b_offset

            # Spike-and-slab priors
            sigma_ind_raw = ed.Normal(
                loc=tf.zeros(shape=[D, K-1], dtype=dtype),
                scale=tf.ones(shape=[D, K-1], dtype=dtype),
                name='sigma_ind_raw')
            ind_t = sigma_ind_raw * 50
            ind = tf.exp(ind_t) / (1 + tf.exp(ind_t))

            # Calculate betas
            beta = ind * b_raw

            # Include slope 0 for baseline cell type
            beta = tf.concat(axis=1, values=[beta[:, :baseline_index],
                                             tf.fill(value=0., dims=[D, 1]),
                                             beta[:, baseline_index:]])

            # Concentration vector from intercepts, slopes
            concentration_ = tf.exp(alpha + tf.matmul(x, beta))

            # Cell count prediction via DirMult
            predictions = ed.DirichletMultinomial(n_total, concentration=concentration_, name="predictions")
            return predictions

        # Joint posterior distribution
        self.log_joint = ed.make_log_joint_fn(define_model)
        # Function to compute log posterior probability

        self.target_log_prob_fn = lambda alpha_, mu_b_, sigma_b_, b_offset_, sigma_ind_raw_:\
            self.log_joint(x=self.x,
                           n_total=self.n_total,
                           K=self.K,
                           predictions=self.y,
                           alpha=alpha_,
                           mu_b=mu_b_,
                           sigma_b=sigma_b_,
                           b_offset=b_offset_,
                           sigma_ind_raw=sigma_ind_raw_,
                           )

        alpha_size = [self.K]
        beta_size = [self.D, self.K-1]

        # MCMC starting values
        self.params = [tf.random.normal(alpha_size, 0, 1, name='init_alpha'),
                       tf.zeros(1, name="init_mu_b", dtype=dtype),
                       tf.ones(1, name="init_sigma_b", dtype=dtype),
                       tf.random.normal(beta_size, 0, 1, name='init_b_offset'),
                       tf.zeros(beta_size, name='init_sigma_ind_raw', dtype=dtype),
                       ]

    # Calculate predicted cell counts (for analysis purposes)
    def get_y_hat(self, states_burnin, num_results, n_burnin):
        """
        Calculate predicted cell counts (for analysis purposes) and add intermediate parameters to MCMC results

        Parameters
        ----------
        states_burnin -- List
            MCMC chain without burnin samples
        num_results -- int
            Chain length (with burnin)
        n_burnin -- int
            Number of burnin samples

        Returns
        -------
        y_mean
            predicted cell counts
        """

        chain_size_beta = [num_results - n_burnin, self.D, self.K]
        chain_size_beta_raw = [num_results - n_burnin, self.D, self.K-1]
        chain_size_y = [num_results - n_burnin, self.N, self.K]

        alphas = states_burnin[0]
        alphas_final = alphas.mean(axis=0)

        ind_raw = states_burnin[4] * 50
        mu_b = states_burnin[1]
        sigma_b = states_burnin[2]
        b_offset = states_burnin[3]

        ind_ = np.exp(ind_raw) / (1 + np.exp(ind_raw))

        b_raw_ = mu_b.reshape((num_results - n_burnin, 1, 1)) + np.einsum("...jk, ...j->...jk", b_offset, sigma_b)

        beta_temp = np.einsum("..., ...", ind_, b_raw_)

        beta_ = np.zeros(chain_size_beta)
        for i in range(num_results - n_burnin):
            beta_[i] = np.concatenate([beta_temp[i, :, :self.baseline_index],
                                       np.zeros(shape=[self.D, 1], dtype=np.float32),
                                       beta_temp[i, :, self.baseline_index:]], axis=1)

        conc_ = np.exp(np.einsum("jk, ...kl->...jl", self.x, beta_)
                       + alphas.reshape((num_results - n_burnin, 1, self.K))).astype(np.float32)

        predictions_ = np.zeros(chain_size_y)
        for i in range(num_results - n_burnin):
            pred = tfd.DirichletMultinomial(self.n_total, conc_[i, :, :]).mean().numpy()
            predictions_[i, :, :] = pred

        betas_final = beta_.mean(axis=0)
        states_burnin.append(ind_)
        states_burnin.append(b_raw_)
        states_burnin.append(beta_)
        states_burnin.append(conc_)
        states_burnin.append(predictions_)

        concentration = np.exp(np.matmul(self.x, betas_final) + alphas_final).astype(np.float32)
        y_mean = concentration / np.sum(concentration, axis=1, keepdims=True) * self.n_total.numpy()[:, np.newaxis]
        return y_mean


class NoBaselineModelNoEdward(CompositionalModel):

    """
    WIP! DO NOT USE!!!

    implements statistical model and
    test statistics for compositional differential change analysis
    without specification of a baseline cell type
    """

    def __init__(self, *args, **kwargs):
        """
        Constructor of model class

        :param covariate_matrix: numpy array [NxD] - covariate matrix
        :param data_matrix: numpy array [NxK] - cell count matrix
        :param sample_counts: numpy array [N] - number of cells per sample
        :param dtype: data type for all numbers (for tensorflow)

        :return
        NotImplementedError
        """

        super(self.__class__, self).__init__(*args, **kwargs)

        self.baseline_index = None
        dtype = tf.float32

        # All parameters that are returned for analysis
        self.param_names = ["mu_b", "sigma_b", "b_offset", "ind_raw", "alpha",
                            "ind", "b_raw", "beta", "concentration", "prediction"]

        alpha_size = [self.K]
        beta_size = [self.D, self.K]

        Root = tfd.JointDistributionCoroutine.Root

        def model():
            mu_b = yield Root(tfd.Independent(
                tfd.Normal(loc=tf.zeros(1, dtype=dtype),
                           scale=tf.ones(1, dtype=dtype),
                           name="mu_b"),
                reinterpreted_batch_ndims=1))

            sigma_b = yield Root(tfd.Independent(
                tfd.HalfCauchy(tf.zeros(1, dtype=dtype),
                               tf.ones(1, dtype=dtype),
                               name="sigma_b"),
                reinterpreted_batch_ndims=1))

            b_offset = yield Root(tfd.Independent(
                tfd.Normal(
                    loc=tf.zeros(beta_size, dtype=dtype),
                    scale=tf.ones(beta_size, dtype=dtype),
                    name="b_offset"),
                reinterpreted_batch_ndims=2))

            # Spike-and-slab
            ind_raw = yield Root(tfd.Independent(
                tfd.Normal(
                    loc=tf.zeros(shape=beta_size, dtype=dtype),
                    scale=tf.ones(shape=beta_size, dtype=dtype),
                    name='ind_raw'),
                reinterpreted_batch_ndims=2))

            ind_scaled = ind_raw * 50
            ind = tf.exp(ind_scaled) / (1 + tf.exp(ind_scaled))

            b_raw = mu_b + sigma_b * b_offset

            beta = ind * b_raw

            alpha = yield Root(tfd.Independent(
                tfd.Normal(
                    loc=tf.zeros(alpha_size),
                    scale=tf.ones(alpha_size) * 5,
                    name="alpha"),
                reinterpreted_batch_ndims=1))

            concentrations = tf.exp(alpha + tf.matmul(tf.cast(self.x, dtype), beta))

            # Cell count prediction via DirMult
            predictions = yield Root(tfd.Independent(
                tfd.DirichletMultinomial(
                    total_count=tf.cast(self.n_total, dtype),
                    concentration=concentrations,
                    name="predictions"),
                reinterpreted_batch_ndims=1))

        self.model_struct = tfd.JointDistributionCoroutine(model)

        # Joint posterior distribution
        self.target_log_prob_fn = lambda *args:\
            self.model_struct.log_prob(list(args) + [tf.cast(self.y, dtype)])

        # MCMC starting values
        self.params = [tf.zeros(1, name="init_mu_b", dtype=dtype),
                       tf.ones(1, name="init_sigma_b", dtype=dtype),
                       tf.zeros(beta_size, name='init_b_offset', dtype=dtype),
                       tf.zeros(beta_size, name='init_ind_raw', dtype=dtype),
                       tf.zeros(alpha_size, name='init_alpha', dtype=dtype)
                       ]

    # Calculate predicted cell counts (for analysis purposes)
    def get_y_hat(self, states_burnin, num_results, n_burnin):
        chain_size_y = [num_results - n_burnin, self.N, self.K]

        alphas = states_burnin[4]
        alphas_final = alphas.mean(axis=0)

        ind_raw = states_burnin[3] * 50
        mu_b = states_burnin[0]
        sigma_b = states_burnin[1]
        b_offset = states_burnin[2]

        ind_ = np.exp(ind_raw) / (1 + np.exp(ind_raw))

        b_raw_ = mu_b.reshape((num_results - n_burnin, 1, 1)) + np.einsum("...jk, ...j->...jk", b_offset, sigma_b)

        beta_ = np.einsum("..., ...", ind_, b_raw_)

        conc_ = np.exp(np.einsum("jk, ...kl->...jl", self.x, beta_)
                       + alphas.reshape((num_results - n_burnin, 1, self.K)))

        predictions_ = np.zeros(chain_size_y)
        for i in range(num_results - n_burnin):
            pred = tfd.DirichletMultinomial(self.n_total, conc_[i, :, :]).mean().numpy()
            predictions_[i, :, :] = pred

        betas_final = beta_.mean(axis=0)
        states_burnin.append(ind_)
        states_burnin.append(b_raw_)
        states_burnin.append(beta_)
        states_burnin.append(conc_)
        states_burnin.append(predictions_)

        concentration = np.exp(np.matmul(self.x, betas_final) + alphas_final).astype(np.float32)

        y_mean = concentration / np.sum(concentration, axis=1, keepdims=True) * self.n_total.numpy()[:, np.newaxis]

        return y_mean


class BaselineModelNoEdward(CompositionalModel):

    """
    WIP! DO NOT USE!!!

    implements statistical model and
    test statistics for compositional differential change analysis
    without specification of a baseline cell type
    """

    def __init__(self, baseline_index, *args, **kwargs):
        """
        Constructor of model class

        :param covariate_matrix: numpy array [NxD] - covariate matrix
        :param data_matrix: numpy array [NxK] - cell count matrix
        :param sample_counts: numpy array [N] - number of cells per sample
        :param dtype: data type for all numbers (for tensorflow)

        :return
        NotImplementedError
        """

        super(self.__class__, self).__init__(*args, **kwargs)

        self.baseline_index = baseline_index
        dtype = tf.float32

        # All parameters that are returned for analysis
        self.param_names = ["mu_b", "sigma_b", "b_offset", "ind_raw", "alpha",
                            "ind", "b_raw", "beta", "concentration", "prediction"]

        alpha_size = [self.K]
        beta_size = [self.D, self.K]
        beta_nobl_size = [self.D, self.K-1]

        Root = tfd.JointDistributionCoroutine.Root

        def model():
            mu_b = yield Root(tfd.Independent(
                tfd.Normal(loc=tf.zeros(1, dtype=dtype),
                           scale=tf.ones(1, dtype=dtype),
                           name="mu_b"),
                reinterpreted_batch_ndims=1))

            sigma_b = yield Root(tfd.Independent(
                tfd.HalfCauchy(tf.zeros(1, dtype=dtype),
                               tf.ones(1, dtype=dtype),
                               name="sigma_b"),
                reinterpreted_batch_ndims=1))

            b_offset = yield Root(tfd.Independent(
                tfd.Normal(
                    loc=tf.zeros(beta_nobl_size, dtype=dtype),
                    scale=tf.ones(beta_nobl_size, dtype=dtype),
                    name="b_offset"),
                reinterpreted_batch_ndims=2))

            # Spike-and-slab
            ind_raw = yield Root(tfd.Independent(
                tfd.Normal(
                    loc=tf.zeros(shape=beta_nobl_size, dtype=dtype),
                    scale=tf.ones(shape=beta_nobl_size, dtype=dtype),
                    name='ind_raw'),
                reinterpreted_batch_ndims=2))

            ind_scaled = ind_raw * 50
            ind = tf.exp(ind_scaled) / (1 + tf.exp(ind_scaled))

            b_raw = mu_b + sigma_b * b_offset

            beta = ind * b_raw

            # Include slope 0 for baseline cell type
            beta = tf.concat(axis=1, values=[beta[:, :baseline_index],
                                             tf.fill(value=0., dims=[self.D, 1]),
                                             beta[:, baseline_index:]])

            alpha = yield Root(tfd.Independent(
                tfd.Normal(
                    loc=tf.zeros(alpha_size),
                    scale=tf.ones(alpha_size) * 5,
                    name="alpha"),
                reinterpreted_batch_ndims=1))

            concentrations = tf.exp(alpha + tf.matmul(self.x, beta))

            # Cell count prediction via DirMult
            predictions = yield Root(tfd.Independent(
                tfd.DirichletMultinomial(
                    total_count=tf.cast(self.n_total, dtype),
                    concentration=concentrations,
                    name="predictions"),
                reinterpreted_batch_ndims=1))

        self.model_struct = tfd.JointDistributionCoroutine(model)

        # Joint posterior distribution
        self.target_log_prob_fn = lambda *args:\
            self.model_struct.log_prob(list(args) + [tf.cast(self.y, dtype)])

        # MCMC starting values
        self.params = [tf.zeros(1, name="init_mu_b", dtype=dtype),
                       tf.ones(1, name="init_sigma_b", dtype=dtype),
                       tf.random.normal(beta_nobl_size, 0, 1, name='init_b_offset', dtype=dtype),
                       tf.zeros(beta_nobl_size, name='init_ind_raw', dtype=dtype),
                       tf.random.normal(alpha_size, 0, 1, name='init_alpha', dtype=dtype)
                       ]

    # Calculate predicted cell counts (for analysis purposes)
    def get_y_hat(self, states_burnin, num_results, n_burnin):
        chain_size_y = [num_results - n_burnin, self.N, self.K]

        alphas = states_burnin[4]
        alphas_final = alphas.mean(axis=0)

        ind_raw = states_burnin[3] * 50
        mu_b = states_burnin[0]
        sigma_b = states_burnin[1]
        b_offset = states_burnin[2]

        ind_ = np.exp(ind_raw) / (1 + np.exp(ind_raw))

        b_raw_ = mu_b.reshape((num_results - n_burnin, 1, 1)) + np.einsum("...jk, ...j->...jk", b_offset, sigma_b)

        beta_ = np.einsum("..., ...", ind_, b_raw_)

        conc_ = np.exp(np.einsum("jk, ...kl->...jl", self.x, beta_)
                       + alphas.reshape((num_results - n_burnin, 1, self.K)))

        predictions_ = np.zeros(chain_size_y)
        for i in range(num_results - n_burnin):
            pred = tfd.DirichletMultinomial(self.n_total, conc_[i, :, :]).mean().numpy()
            predictions_[i, :, :] = pred

        betas_final = beta_.mean(axis=0)
        states_burnin.append(ind_)
        states_burnin.append(b_raw_)
        states_burnin.append(beta_)
        states_burnin.append(conc_)
        states_burnin.append(predictions_)

        concentration = np.exp(np.matmul(self.x, betas_final) + alphas_final).astype(np.float32)

        y_mean = concentration / np.sum(concentration, axis=1, keepdims=True) * self.n_total.numpy()[:, np.newaxis]

        return y_mean



"""
Functions for plotting and analyzing the results from running models over multiple parameter sets

:authors: Johannes Ostner
"""


import numpy as np
import seaborn as sns
import pandas as pd
import pickle as pkl
import os
import ast
import matplotlib.pyplot as plt

# Helpers for loading result classes from old environment
import io


class RenameUnpickler(pkl.Unpickler):
    def find_class(self, module, name):
        renamed_module = module
        if module == "multi_parameter_sampling" or module == "model.multi_parameter_sampling":
            renamed_module = "scCODA.util.multi_parameter_sampling"
        if module == "compositional_analysis_generation_toolbox" or module == "model.compositional_analysis_generation_toolbox":
            renamed_module = "scCODA.util.data_generation"
        if module == "result_classes" or module == "model.result_classes":
            renamed_module = "scCODA.util.result_classes"
        if module == "final_models" or module == "model.final_models":
            renamed_module = "scCODA.model.dirichlet_models"
        if module == "SCDCpy.util":
            renamed_module = "scCODA.util"

        renamed_name = name
        if name == 'Multi_param_simulation':
            renamed_name = "MultiParamSimulation"
        if name == 'MCMCResult':
            renamed_name = "CAResult"

        return super(RenameUnpickler, self).find_class(renamed_module, renamed_name)


def renamed_load(file_obj):
    return RenameUnpickler(file_obj).load()


def renamed_loads(pickled_bytes):
    file_obj = io.BytesIO(pickled_bytes)
    return renamed_load(file_obj)


def multi_run_study_analysis_prepare(path, file_identifier="result_", custom_threshold=None):

    """
    Function to read in and calculate discovery rates, ... for an entire directory of multi_parameter_sampling files

    Parameters
    ----------
    path -- str
        path to directory
    file_identifier -- str (optional)
        identifier that is part of all files we want to analyze - only these files are loaded
    custom_threshold -- float (optional)
        custom spike-and-slab threshold

    Returns
    -------
    results -- list
        List of raw result files
    all_study_params -- pandas DataFrame
        Parameters and result data for all files
    all_study_params_agg -- pandas DataFrame
        Parameters and result data, aggregated over all files with identical parameters

    """

    files = os.listdir(path)

    results = []

    print("Calculating discovery rates...")
    i = 0

    # For all files:
    for f in files:
        i += 1

        print("Preparing: ", i / len(files))
        if file_identifier in f:
            # Load file
            r = renamed_load(open(path + "/" + f, "rb"))

            if custom_threshold is not None:
                for r_k, r_i in r.mcmc_results.items():
                    r.mcmc_results[r_k].params["final_parameter"] = np.where(np.isnan(r_i.params["mean_nonzero"]),
                                                                             r_i.params["mean"],
                                                                             np.where(r_i.params[
                                                                                          "inclusion_prob"] > custom_threshold,
                                                                                      r_i.params["mean_nonzero"],
                                                                                      0))

            # Discovery rates for beta
            r.get_discovery_rates()

            results.append(r)

    # Generate all_study_params
    all_study_params = pd.concat([r.parameters for r in results])
    simulation_parameters = ["cases", "K", "n_total", "n_samples", "b_true", "w_true", "num_results"]
    all_study_params[simulation_parameters] = all_study_params[simulation_parameters].astype(str)

    # Aggregate over identical parameter sets
    all_study_params_agg = all_study_params.groupby(simulation_parameters).sum()

    return results, all_study_params, all_study_params_agg.reset_index()


def get_scores(agg_df, models=1):
    """
    Calculates extended summary statistics, such as TPR, TNR, youden index, f1-score, MCC

    Parameters
    ----------
    agg_df -- pandas DataFrame
    models -- int
        number of different models used for the analysis (relevant if data comes from MultiParameterSamplingMultiModel)

    Returns
    -------
    agg_df -- pandas DataFrame
        Same as input, with added columns for summary statistics

    """

    if models == 1:
        tp = agg_df["tp"]
        tn = agg_df["tn"]
        fp = agg_df["fp"]
        fn = agg_df["fn"]

        tpr = (tp / (tp + fn)).fillna(0)
        agg_df["tpr"] = tpr
        tnr = (tn / (tn + fp)).fillna(0)
        agg_df["tnr"] = tnr
        precision = (tp / (tp + fp)).fillna(0)
        agg_df["precision"] = precision
        acc = (tp + tn) / (tp + tn + fp + fn).fillna(0)
        agg_df["accuracy"] = acc

        agg_df["youden"] = tpr + tnr - 1
        agg_df["f1_score"] = 2 * (tpr * precision / (tpr + precision)).fillna(0)

        agg_df["mcc"] = (((tp * tn) - (fp * fn)) / np.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn))).fillna(0)

    else:
        for m in range(models):
            m_str = str(m)
            tp = agg_df["tp_" + m_str]
            tn = agg_df["tn_" + m_str]
            fp = agg_df["fp_" + m_str]
            fn = agg_df["fn_" + m_str]

            tpr = (tp / (tp + fn)).fillna(0)
            agg_df["tpr_" + m_str] = tpr
            tnr = (tn / (tn + fp)).fillna(0)
            agg_df["tnr_" + m_str] = tnr
            precision = (tp / (tp + fp)).fillna(0)
            agg_df["precision_" + m_str] = precision
            acc = (tp + tn) / (tp + tn + fp + fn).fillna(0)
            agg_df["accuracy_" + m_str] = acc

            agg_df["youden_" + m_str] = tpr + tnr - 1
            agg_df["f1_score_" + m_str] = 2 * (tpr * precision / (tpr + precision)).fillna(0)

            agg_df["mcc_" + m_str] = (((tp * tn) - (fp * fn)) / np.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn))).fillna(0)

    return agg_df


def plot_discovery_rates_agg(rates_df, dim_1='w_true', dim_2=None, path=None):
    """
    Plot heatmap of TPR and TNR for one parameter series vs. another


    Parameters
    ----------
    rates_df -- pandas DataFrame
        format as all_study_params_agg from multi_run_study_analysis_prepare
    dim_1 -- str
        parameter on x-axis
    dim_2 -- str
        parameter on y-axis
    path -- str
        directory to save plot to

    Returns
    -------

    """

    sns.set_style("whitegrid")
    fig, ax = plt.subplots(1, 2, figsize=(13, 5))

    # If no second dimension specified, make a 1d-heatmap
    if dim_2 is None:
        rates_df = rates_df.groupby(dim_1).mean().reset_index()
        dim_2 = "x"
        # Generate dataframe for plotting
        plot_data = pd.DataFrame({dim_1: rates_df[dim_1],
                                  dim_2: [1 for i in range(rates_df.shape[0])],
                                  "tpr": rates_df["tpr"],
                                  "tnr": rates_df["tnr"]
                                  })
        # plot Heatmaps
        sns.heatmap(plot_data.pivot(dim_1, dim_2, 'tpr'), ax=ax[0], vmin=0, vmax=1).set_title("MCMC TPR")
        sns.heatmap(plot_data.pivot(dim_1, dim_2, 'tnr'), ax=ax[1], vmin=0, vmax=1).set_title("MCMC TNR")
    else:
        rates_df = rates_df.groupby([dim_1, dim_2]).mean().reset_index()
        # Generate dataframe for plotting
        plot_data = pd.DataFrame({dim_1: rates_df[dim_1],
                                  dim_2: rates_df[dim_2],
                                  "tpr": rates_df["tpr"],
                                  "tnr": rates_df["tnr"]
                                  })
        # plot Heatmaps
        sns.heatmap(plot_data.pivot(dim_1, dim_2, 'tpr'), ax=ax[0], vmin=0, vmax=1).set_title("MCMC TPR")
        sns.heatmap(plot_data.pivot(dim_1, dim_2, 'tnr'), ax=ax[1], vmin=0, vmax=1).set_title("MCMC TNR")

    # Save
    if path is not None:
        plt.savefig(path)

    plt.show()


def plot_cases_vs_controls(rates_df, results, identifier_w="", path=None, suptitle=None):
    """
    Plot heatmaps of TPR and TNR for number of case vs. number of control samples.


    Parameters
    ----------
    rates_df -- pandas DataFrame
        Same as all_study_params_agg from multi_run_study_analysis_prepare
    results -- pandas DataFrame
        same format as results from multi_run_study_analysis_prepare
    identifier_w -- str (optional)
        if plotting only a subset of all ground truth effects
    path -- str
        directory to save plot to
    suptitle -- str
        Header for entirety of plots

    Returns
    -------

    """

    sns.set_style("whitegrid")
    fig, ax = plt.subplots(2, 2, figsize=(12, 10))

    # DataFrame for heatmaps
    rates_df = rates_df.loc[(rates_df["w_true"] == str(identifier_w))].groupby("n_samples").mean().reset_index()
    plot_data = pd.DataFrame({"controls": [ast.literal_eval(x)[0] for x in rates_df["n_samples"].tolist()],
                              "cases": [ast.literal_eval(x)[1] for x in rates_df["n_samples"].tolist()],
                              "tpr": rates_df["tpr"],
                              "tnr": rates_df["tnr"]
                              })

    # Plot heatmaps
    sns.heatmap(plot_data.pivot("controls", "cases", 'tpr'), ax=ax[0, 0], vmin=0, vmax=1).set_title("MCMC TPR")
    sns.heatmap(plot_data.pivot("controls", "cases", 'tnr'), ax=ax[0, 1], vmin=0, vmax=1).set_title("MCMC TNR")

    # Case vs. control count boxplots
    cases_y = []
    controls_y = []

    # Get count data
    for r in results:
        if r.parameters.loc[0, "w_true"] == identifier_w:
            for i in r.mcmc_results:
                n_cases = r.parameters.loc[i, "n_samples"][0]

                controls_y.extend(r.mcmc_results[i].y[:n_cases].tolist())
                cases_y.extend(r.mcmc_results[i].y[n_cases:].tolist())

    cases_y = pd.DataFrame(cases_y)
    controls_y = pd.DataFrame(controls_y)

    # Generate boxplots
    sns.boxplot(data=controls_y, ax=ax[1, 0]).set_title("control group cell counts")
    sns.boxplot(data=cases_y, ax=ax[1, 1]).set_title("case group cell counts")

    # Add title
    if suptitle is not None:
        plt.suptitle(suptitle)
    plt.tight_layout()

    # Save plot
    if path is not None:
        plt.savefig(path)

    plt.show()


def multi_run_study_analysis_prepare_per_param(path, file_identifier="result_"):
    """
    Function to calculate discovery rates, ... for an entire directory of multi_parameter_sampling files.
    Effect Discovery rates are calculated separately for each cell type.

    Parameters
    ----------
    path -- str
        path to directory
    file_identifier -- str (optional)
        identifier that is part of all files we want to analyze

    Returns
    -------
    results -- list
        List of raw result files
    all_study_params -- pandas DataFrame
        Parameters and result data for all files
    all_study_params_agg -- pandas DataFrame
        Parameters and result data, aggregated over all sets of identical parameters
    """

    files = os.listdir(path)

    results = []

    print("Calculating discovery rates...")
    i = 0

    for f in files:
        i += 1

        print("Preparing: ", i / len(files))
        if file_identifier in f:
            # Load file
            r = renamed_load(open(path + "/" + f, "rb"))

            # Discovery rates for beta per parameter
            r.get_discovery_rates_per_param()

            results.append(r)

    # Generate all_study_params
    all_study_params = pd.concat([r.parameters for r in results])
    simulation_parameters = ["cases", "K", "n_total", "n_samples", "b_true", "w_true", "num_results"]
    all_study_params[simulation_parameters] = all_study_params[simulation_parameters].astype(str)

    # Aggregate over identical parameter sets
    all_study_params_agg = all_study_params.groupby(simulation_parameters).sum()

    return results, all_study_params, all_study_params_agg.reset_index()


def plot_cases_vs_controls_per_param(K, rates_df, results, identifier_w, path=None, suptitle=None):
    """
    Plot heatmaps of discovery rate for number of case vs. number of control samples, for each cell type.
    Also plots counts of cases vs. controls for all cell types

    Parameters
    ----------
    K -- int

    rates_df -- pandas DataFrame
        same format as all_study_params_agg from multi_run_study_analysis_prepare
    results -- pandas DataFrame
        same format as results from multi_run_study_analysis_prepare
    identifier_w -- str
        if plotting only a subset of all ground truth effects
    path -- str
        directory to save plot to
    suptitle -- str
        Header for entirety of plots

    Returns
    -------

    """

    # plot initialization
    sns.set_style("whitegrid")
    fig, ax = plt.subplots(2, K, figsize=(K*6, 10))
    for a in ax[1, :]:
        a.set_ylim(0, 1000)

    # Get heatmap data relevant for plotting
    rates_df = rates_df.loc[(rates_df["w_true"] == str(identifier_w))].groupby("n_samples").mean().reset_index()
    rates_df["controls"] = [ast.literal_eval(x)[0] for x in rates_df["n_samples"].tolist()]
    rates_df["cases"] = [ast.literal_eval(x)[1] for x in rates_df["n_samples"].tolist()]

    # For each cell type:
    for i in range(K):
        plot_df = rates_df.loc[:, ["controls", "cases", "correct_"+str(i), "false_"+str(i)]]
        plot_df["disc_rate"] = plot_df["correct_"+str(i)]/(plot_df["correct_"+str(i)] + plot_df["false_"+str(i)])
        # If Effect on cell type ==0: plot in blue, else plot in red
        if identifier_w[0][i] == 0:
            cmap = "Blues_r"
        else:
            cmap = "Reds_r"
        # Plot heatmap
        sns.heatmap(plot_df.pivot("controls", "cases", 'disc_rate'), ax=ax[0, i], vmin=0, vmax=1, cmap=cmap).\
            set_title("Cell type "+str(i+1)+" accuracy - " + "Effect: " + str(identifier_w[0][i]))

    # Get cell count data for each cell type
    cases_y = []
    controls_y = []
    for r in results:
        if r.parameters.loc[0, "w_true"] == identifier_w:
            for i in r.mcmc_results:
                n_cases = r.parameters.loc[i, "n_samples"][0]

                controls_y.extend(r.mcmc_results[i].y[:n_cases].tolist())
                cases_y.extend(r.mcmc_results[i].y[n_cases:].tolist())
    cases_y = cases_y
    controls_y = controls_y

    # Plot boxplots
    for i in range(K):
        box_df = pd.DataFrame({"controls": [y[i] for y in controls_y],
                               "cases": [y[i] for y in cases_y]})
        lf_change = np.round(np.log2(box_df["cases"].mean() / box_df["controls"].mean()), 2)
        sns.boxplot(data=box_df.loc[:, ["cases", "controls"]], ax=ax[1, i], order=["controls", "cases"]).\
            set_title("Log-fold change: "+str(lf_change))

    # Add title
    if suptitle is not None:
        plt.suptitle(suptitle)

    # Save plot
    plt.tight_layout()
    if path is not None:
        plt.savefig(path)

    plt.show()


def multi_run_study_analysis_multi_model_prepare(path, file_identifier="result_"):
    """
    Function to calculate discovery rates, ... for an entire directory of multi_parameter_sampling_multi_model files

    Parameters
    ----------
    path -- str
        path to directory
    file_identifier -- str (optional)
        identifier that is part of all files we want to analyze

    Returns
    -------
    results -- list
        List of raw result files
    all_study_params -- pandas DataFrame
        Parameters and result data for all files
    all_study_params_agg -- pandas DataFrame
        Parameters and result data, aggregated over all sets of identical parameters
    """

    files = os.listdir(path)
    results = []
    print("Calculating discovery rates...")
    i = 0

    # For all files:
    for f in files:
        i += 1

        print("Preparing: ", i / len(files))
        if file_identifier in f:
            # Load file
            r = renamed_load(open(path + "/" + f, "rb"))

            # Discovery rates for beta
            r.get_discovery_rates()

            results.append(r)

    # Generate all_study_params
    all_study_params = pd.concat([r.parameters for r in results])
    simulation_parameters = ["cases", "K", "n_total", "n_samples", "b_true", "w_true", "num_results"]
    all_study_params[simulation_parameters] = all_study_params[simulation_parameters].astype(str)

    # Aggregate over identical parameter sets
    all_study_params_agg = all_study_params.groupby(simulation_parameters).sum()

    return results, all_study_params, all_study_params_agg.reset_index()


def plot_cases_vs_controls_per_param_2(K, rates_df, results, identifier_w, path=None, suptitle=None):
    """
    Optimized version of plot_cases_vs_controls_per_param

    Parameters
    ----------
    K -- int
        number of cell types
    rates_df -- pandas DataFrame
        same format as all_study_params_agg from multi_run_study_analysis_prepare
    results -- pandas DataFrame
        same format as results from multi_run_study_analysis_prepare
    identifier_w -- str (optional)
        if plotting only a subset of all ground truth effects
    path -- str
        directory to save plot to
    suptitle -- str
        Header for entirety of plots

    Returns
    -------

    """


    sns.set_style("whitegrid")

    # Get heatmap data relevant for plotting
    rates_df = rates_df.loc[(rates_df["w_true"] == str(identifier_w))].groupby("n_samples").mean().reset_index()
    rates_df["controls"] = [ast.literal_eval(x)[0] for x in rates_df["n_samples"].tolist()]
    rates_df["cases"] = [ast.literal_eval(x)[1] for x in rates_df["n_samples"].tolist()]

    # Get cell count data for each cell type
    cases_y = []
    controls_y = []
    for r in results:
        if r.parameters.loc[0, "w_true"] == identifier_w:
            for i in r.mcmc_results:
                n_cases = r.parameters.loc[i, "n_samples"][0]

                controls_y.extend(r.mcmc_results[i].y[:n_cases].tolist())
                cases_y.extend(r.mcmc_results[i].y[n_cases:].tolist())
    cases_y = cases_y
    controls_y = controls_y

    # For each cell type:
    for i in range(K):
        # Initialize plot
        fig, ax = plt.subplots(2, 1, figsize=(6, 10))
        ax[1].set_ylim(0, 1000)

        # DataFrame for heatmap
        plot_df = rates_df.loc[:, ["controls", "cases", "correct_"+str(i), "false_"+str(i)]]
        plot_df["disc_rate"] = plot_df["correct_"+str(i)]/(plot_df["correct_"+str(i)] + plot_df["false_"+str(i)])
        # If Effect on cell type ==0: plot in blue, else plot in red
        if identifier_w[0][i] == 0:
            cmap = "Blues_r"
        else:
            cmap = "Reds_r"
        # Plot heatmap
        sns.heatmap(plot_df.pivot("controls", "cases", 'disc_rate'), ax=ax[0], vmin=0, vmax=1, cmap=cmap).\
            set_title("Accuracy")

        # DataFrame for boxplot
        box_df = pd.DataFrame({"controls": [y[i] for y in controls_y],
                               "cases": [y[i] for y in cases_y]})
        change = np.round(box_df["cases"].mean() - box_df["controls"].mean(), 2)
        sns.boxplot(data=box_df.loc[:, ["cases", "controls"]], ax=ax[1], order=["controls", "cases"]).\
            set_title("Average change: "+str(change)+" cells")

        # Add title
        if suptitle is not None:
            plt.suptitle(suptitle)

        # Save plot
        plt.tight_layout()
        if path is not None:
            plt.savefig(path + "_type_" + str(i+1).replace(".", ""), bbox_inches="tight")

    plt.show()



"""
Results class that summarizes the results of the different inference methods and calculates test statistics

:authors: Johannes Ostner
"""
import numpy as np
import arviz as az
import pandas as pd
import pickle as pkl


class CAResultConverter(az.data.io_dict.DictConverter):
    """
    Helper class for result conversion in arviz's format
    """

    def to_result_data(self, sampling_stats, model_specs):

        post = self.posterior_to_xarray()
        ss = self.sample_stats_to_xarray()
        postp = self.posterior_predictive_to_xarray()
        prior = self.prior_to_xarray()
        ssp = self.sample_stats_prior_to_xarray()
        prip = self.prior_predictive_to_xarray()
        obs = self.observed_data_to_xarray()

        return CAResult(
            sampling_stats, model_specs,
            **{
                "posterior": post,
                "sample_stats": ss,
                "posterior_predictive": postp,
                "prior": prior,
                "sample_stats_prior": ssp,
                "prior_predictive": prip,
                "observed_data": obs,
            }
        )


class CAResult(az.InferenceData):
    """
    Result class, extends the arviz framework for inference data.

    The CAResult class is an extension of az.InferenceData, that adds some information about the compositional model.
    It supports all functionality from az.InferenceData.
    """

    def __init__(self, sampling_stats, model_specs, **kwargs):
        """
        The following attributes are added during class initialization:

        self.sampling_stats: dict - see below
        self.model_specs: dict - see below

        self.intercept_df: Summary dataframe from CAResult.summary_prepare
        self.effect_df: Summary dataframe from CAResult.summary_prepare

        Parameters
        ----------
        sampling_stats -- dict
            Information and statistics about the MCMC sampling procedure.
            Default keys:
            "chain_length": Length of MCMC chain (with burnin samples)
            "n_burnin": Number of burnin samples
            "acc_rate": MCMC Acceptance rate
            "duration": Duration of MCMC sampling
        model_specs -- dict
            All information and statistics about the model specifications.
            Default keys:
            "formula": Formula string
            "baseline": int - identifier of baseline cell type
            Added during class initialization: "threshold_prob": Threshold for inclusion probability that separates significant from non-significant effects
        kwargs -- passed to az.InferenceData
        """
        super(self.__class__, self).__init__(**kwargs)

        self.sampling_stats = sampling_stats
        self.model_specs = model_specs

        intercept_df, effect_df = self.summary_prepare()

        self.intercept_df = intercept_df
        self.effect_df = effect_df

    def summary_prepare(self, *args, **kwargs):
        """
        Generates summary dataframes for intercepts and slopes.
        This function supports all functionalities from az.summary.

        Parameters
        ----------
        args -- Passed to az.summary
        kwargs -- Passed to az.summary

        Returns
        -------
        intercept_df -- pandas df
            Summary of intercept parameters. Contains one row per cell type. Columns:
            Final Parameter: Final intercept model parameter
            hdi X%: Upper and lower boundaries of confidence interval (width specified via hdi_prob=)
            SD: Standard deviation of MCMC samples
            Expected sample: Expected cell counts for a sample with no present covariates. See the tutorial for more explanation

        effect_df -- pandas df
            Summary of effect (slope) parameters. Contains one row per covariate/cell type combination. Columns:
            Final Parameter: Final effect model parameter. If this parameter is 0, the effect is not significant, else it is.
            HDI X%: Upper and lower boundaries of confidence interval (width specified via hdi_prob=)
            SD: Standard deviation of MCMC samples
            Expected sample: Expected cell counts for a sample with only the current covariate set to 1. See the tutorial for more explanation
            log2-fold change: Log2-fold change between expected cell counts with no covariates and with only the current covariate
            Inclusion probability: Share of MCMC samples, for which this effect was not set to 0 by the spike-and-slab prior.
        """
        # initialize summary df
        summ = az.summary(self, *args, **kwargs, kind="stats", var_names=["alpha", "beta"])
        effect_df = summ.loc[summ.index.str.match("|".join(["beta"]))].copy()
        intercept_df = summ.loc[summ.index.str.match("|".join(["alpha"]))].copy()

        # Build neat index
        cell_types = self.posterior.coords["cell_type"].values
        covariates = self.posterior.coords["covariate"].values

        intercept_df.index = pd.Index(cell_types, name="Cell Type")
        effect_df.index = pd.MultiIndex.from_product([covariates, cell_types],
                                                     names=["Covariate", "Cell Type"])

        # Calculation of columns that are not from az.summary
        intercept_df = self.complete_alpha_df(intercept_df)
        effect_df = self.complete_beta_df(intercept_df, effect_df)

        # Give nice column names, remove unnecessary columns
        hdis = intercept_df.columns[intercept_df.columns.str.contains("hdi")]
        hdis_new = hdis.str.replace("hdi_", "HDI ")

        intercept_df = intercept_df.loc[:, ["final_parameter", hdis[0], hdis[1], "sd", "expected_sample"]].copy()
        intercept_df = intercept_df.rename(columns=dict(zip(
            intercept_df.columns,
            ["Final Parameter", hdis_new[0], hdis_new[1], "SD", "Expected Sample"]
        )))

        effect_df = effect_df.loc[:, ["final_parameter", hdis[0], hdis[1], "sd", "inclusion_prob",
                                       "expected_sample", "log_fold"]].copy()
        effect_df = effect_df.rename(columns=dict(zip(
            effect_df.columns,
            ["Final Parameter", hdis_new[0], hdis_new[1], "SD", "Inclusion probability",
             "Expected Sample", "log2-fold change"]
        )))

        return intercept_df, effect_df

    def complete_beta_df(self, intercept_df, effect_df):
        """
        Evaluation of MCMC results for effect parameters. This function is only used within self.summary_prepare.

        Parameters
        ----------
        intercept_df -- Data frame
            Intercept summary, see self.summary_prepare
        effect_df -- Data frame
            Effect summary, see self.summary_prepare

        Returns
        -------
        effect_df -- DataFrame
            DataFrame with inclusion probability, final parameters, expected sample
        """
        beta_inc_prob = []
        beta_nonzero_mean = []

        beta_raw = np.array(self.posterior["beta"])[0]

        # Calculate inclusion prob, nonzero mean for every effect
        for j in range(beta_raw.shape[1]):
            for i in range(beta_raw.shape[2]):
                beta_i_raw = beta_raw[:, j, i]
                beta_i_raw_nonzero = np.where(np.abs(beta_i_raw) > 1e-3)[0]
                prob = beta_i_raw_nonzero.shape[0] / beta_i_raw.shape[0]
                beta_inc_prob.append(prob)
                beta_nonzero_mean.append(beta_i_raw[beta_i_raw_nonzero].mean())

        effect_df.loc[:, "inclusion_prob"] = beta_inc_prob
        effect_df.loc[:, "mean_nonzero"] = beta_nonzero_mean

        # Inclusion prob threshold value
        threshold = 1-(0.77/(beta_raw.shape[2]**0.29))
        self.model_specs["threshold_prob"] = threshold

        # Decide whether betas are significant or not
        effect_df.loc[:, "final_parameter"] = np.where(effect_df.loc[:, "inclusion_prob"] > threshold,
                                                      effect_df.loc[:, "mean_nonzero"],
                                                      0)

        # Get expected sample, log-fold change
        D = len(effect_df.index.levels[0])
        K = len(effect_df.index.levels[1])

        y_bar = np.mean(np.sum(np.array(self.observed_data.y), axis=1))
        alpha_par = intercept_df.loc[:, "final_parameter"]
        alphas_exp = np.exp(alpha_par)
        alpha_sample = (alphas_exp / np.sum(alphas_exp) * y_bar).values

        beta_mean = alpha_par
        beta_sample = []
        log_sample = []

        for d in range(D):
            beta_d = effect_df.loc[:, "final_parameter"].values[(d*K):((d+1)*K)]
            beta_d = (beta_mean + beta_d)
            beta_d = np.exp(beta_d)
            beta_d = beta_d / np.sum(beta_d) * y_bar

            beta_sample = np.append(beta_sample, beta_d)
            log_sample = np.append(log_sample, np.log2(beta_d/alpha_sample))

        effect_df.loc[:, "expected_sample"] = beta_sample
        effect_df.loc[:, "log_fold"] = log_sample

        return effect_df

    def complete_alpha_df(self, intercept_df):
        """
        Evaluation of MCMC results for intercepts. This function is only used within self.summary_prepare.

        Parameters
        ----------
        intercept_df -- Data frame
            Intercept summary, see self.summary_prepare

        Returns
        -------
        intercept_df -- DataFrame
            Summary DataFrame with expected sample, final parameters
        """

        intercept_df = intercept_df.rename(columns={"mean": "final_parameter"})

        # Get expected sample
        y_bar = np.mean(np.sum(np.array(self.observed_data.y), axis=1))
        alphas_exp = np.exp(intercept_df.loc[:, "final_parameter"])
        alpha_sample = (alphas_exp / np.sum(alphas_exp) * y_bar).values
        intercept_df.loc[:, "expected_sample"] = alpha_sample

        return intercept_df

    def summary(self, *args, **kwargs):
        """
        Printing method for summary data

        Usage: result.summary()

        Parameters
        ----------
        args -- Passed to az.summary
        kwargs -- Passed to az.summary

        Returns
        -------

        """

        # If other than default values for e.g. confidence interval are specified, recalculate the intercept and effect DataFrames
        if args or kwargs:
            intercept_df, effect_df = self.summary_prepare(*args, **kwargs)
        else:
            intercept_df = self.intercept_df
            effect_df = self.effect_df

        # Get number of samples, cell types
        if self.sampling_stats["y_hat"] is not None:
            data_dims = self.sampling_stats["y_hat"].shape
        else:
            data_dims = (10, 5)

        # Cut down DataFrames to relevant info
        alphas_print = intercept_df.loc[:, ["Final Parameter", "Expected Sample"]]
        betas_print = effect_df.loc[:, ["Final Parameter", "Expected Sample", "log2-fold change"]]

        # Print everything
        print("Compositional Analysis summary:")
        print("")
        print("Data: %d samples, %d cell types" % data_dims)
        print("Baseline index: %s" % str(self.model_specs["baseline"]))
        print("Formula: %s" % self.model_specs["formula"])
        print("")
        print("Intercepts:")
        print(alphas_print)
        print("")
        print("")
        print("Effects:")
        print(betas_print)

    def summary_extended(self, *args, **kwargs):

        # If other than default values for e.g. confidence interval are specified, recalculate the intercept and effect DataFrames
        if args or kwargs:
            intercept_df, effect_df = self.summary_prepare(*args, **kwargs)
        else:
            intercept_df = self.intercept_df
            effect_df = self.effect_df

        # Get number of samples, cell types
        data_dims = self.sampling_stats["y_hat"].shape

        # Print everything
        print("Compositional Analysis summary (extended):")
        print("")
        print("Data: %d samples, %d cell types" % data_dims)
        print("Baseline index: %s" % str(self.model_specs["baseline"]))
        print("Formula: %s" % self.model_specs["formula"])
        print("Spike-and-slab threshold: {threshold:.3f}".format(threshold=self.model_specs["threshold_prob"]))
        print("")
        print("MCMC Sampling: Sampled {num_results} chain states ({n_burnin} burnin samples) in {duration:.3f} sec. "
              "Acceptance rate: {ar:.1f}%".format(num_results=self.sampling_stats["chain_length"],
                                                  n_burnin=self.sampling_stats["n_burnin"],
                                                  duration=self.sampling_stats["duration"],
                                                  ar=(100*self.sampling_stats["acc_rate"])))
        print("")
        print("Intercepts:")
        print(intercept_df)
        print("")
        print("")
        print("Effects:")
        print(effect_df)

    def compare_to_truth(self, b_true, w_true, *args, **kwargs):
        """
        Extends data frames from summary_prepare by a comparison to some ground truth slope and intercept values

        Parameters
        ----------
        b_true -- pandas Series
            Ground truth slope values. Length must be same as number of cell types
        w_true -- pandas Series
            Ground truth intercept values. Length must be same as number of cell types*number of covariates
        args -- Passed to az.summary
        kwargs -- Passed to az.summary

        Returns
        -------
        intercept_df -- DataFrame
            Summary DataFrame for intercepts
        effect_df -- DataFrame
            Summary DataFrame for effects
        """

        intercept_df, effect_df = self.summary_prepare(*args, **kwargs)

        intercept_df.columns = intercept_df.columns.str.replace('final_parameter', 'predicted')
        effect_df.columns = effect_df.columns.str.replace('final_parameter', 'predicted')

        # Get true params, join to calculated parameters
        b_true = b_true.rename("truth")
        intercept_df = intercept_df.join(b_true)
        w_true = w_true.rename("truth")
        effect_df = effect_df.join(w_true)

        # decide whether effects are found correctly
        intercept_df['dist_to_truth'] = intercept_df['truth'] - intercept_df['predicted']
        intercept_df['effect_correct'] = ((intercept_df['truth'] == 0) == (intercept_df['predicted'] == 0))
        effect_df['dist_to_truth'] = effect_df['truth'] - effect_df['predicted']
        effect_df['effect_correct'] = ((effect_df['truth'] == 0) == (effect_df['predicted'] == 0))

        return intercept_df, effect_df

    def distances(self):
        """
        Compares real cell count matrix to the cell count matrix that arises from the calculated parameters

        Returns
        -------
        ret: DataFrame
        """

        # Get absolute (counts) and relative error matrices
        y = np.array(self.observed_data.y)
        y_hat = self.sampling_stats["y_hat"]
        err = np.abs(y_hat - y)

        err_rel = err / y
        err_rel[np.isinf(err_rel)] = 1.
        err_rel[np.isnan(err_rel)] = 0.

        # Calculate mean errors for each cell type and overall
        avg_abs_cell_type_error = np.mean(err, axis=0, dtype=np.float64)
        avg_rel_cell_type_error = np.mean(err_rel, axis=0, dtype=np.float64)
        avg_abs_total_error = np.mean(err, dtype=np.float64)
        avg_rel_total_error = np.mean(err_rel, dtype=np.float64)

        ret = pd.DataFrame({'Cell Type': np.arange(y.shape[1] + 1),
                            'Absolute Error': np.append(avg_abs_total_error, avg_abs_cell_type_error),
                            'Relative Error': np.append(avg_rel_total_error, avg_rel_cell_type_error),
                            'Actual Means': np.append(np.mean(y, axis=(0, 1)), np.mean(y, axis=0)),
                            'Predicted Means': np.append(np.mean(y_hat, axis=(0, 1)), np.mean(y_hat, axis=0))})

        ret['Cell Type'][0] = 'Total'
        return ret

    def save(self, path_to_file):
        with open(path_to_file, "wb") as f:
            pkl.dump(self, file=f, protocol=4)




"""
Models for the model comparison part in the paper

These models are otherwise not part of scCODA

:authors: Johannes Ostner
"""
import numpy as np
import pandas as pd

import tensorflow as tf
import tensorflow_probability as tfp
import skbio
from tensorflow_probability.python.experimental import edward2 as ed

import statsmodels as sm
from statsmodels.formula.api import glm
from scipy import stats

from sccoda.util import result_classes as res
from sccoda.model import dirichlet_models as dm

tfd = tfp.distributions
tfb = tfp.bijectors

import os

class SimpleModel(dm.CompositionalModel):
    """
    Simple Dirichlet-Multinomial model with normal priors

    """

    def __init__(self, baseline_index, *args, **kwargs):

        super(self.__class__, self).__init__(*args, **kwargs)
        self.baseline_index = baseline_index
        dtype = tf.float32

        # All parameters that are returned for analysis
        self.param_names = ["alpha", "b", "beta", "concentration", "prediction"]

        # Model definition
        def define_model(x, n_total, K):
            N, D = x.shape
            dtype = tf.float32

            alpha = ed.Normal(loc=tf.zeros([K], dtype=dtype), scale=tf.ones([K], dtype=dtype), name="alpha")
            b = ed.Normal(loc=tf.zeros([D, K-1], dtype=dtype), scale=tf.ones([D,K-1], dtype=dtype), name="b")

            beta = tf.concat(axis=1, values=[b[:, :baseline_index],
                                             tf.fill(value=0., dims=[D, 1]),
                                             b[:, baseline_index:]])

            concentration_ = tf.exp(alpha + tf.matmul(x, beta))

            # Likelihood
            predictions = ed.DirichletMultinomial(n_total, concentration=concentration_, name="predictions")
            return predictions

        # Joint posterior distribution
        self.log_joint = ed.make_log_joint_fn(define_model)
        # Function to compute log posterior probability

        self.target_log_prob_fn = lambda alpha_, b_: \
            self.log_joint(x=self.x,
                           n_total=self.n_total,
                           K=self.K,
                           predictions=self.y,
                           alpha=alpha_,
                           b=b_,
                           )

        alpha_size = [self.K]
        beta_size = [self.D, self.K-1]

        self.params = [tf.random.normal(mean=0, stddev=1, name="init_alpha", shape=alpha_size, dtype=dtype),
                       tf.random.normal(mean=0, stddev=1, name="init_b", shape=beta_size, dtype=dtype),
                       ]

    def sample_hmc(self, num_results=int(20e3), n_burnin=int(5e3), num_leapfrog_steps=10, step_size=0.01, num_adapt_steps=None):
        """
        HMC sampling

        Parameters
        ----------
        num_results -- int
            MCMC chain length (default 20000)
        n_burnin -- int
            Number of burnin iterations (default 5000)
        num_leapfrog_steps -- int
            HMC leapfrog steps (default 10)
        step_size -- float
            Initial step size (default 0.01)

        Returns
        -------
        result
            scCODA.util.result_data object
        """

        # (not in use atm)
        constraining_bijectors = [
            tfb.Identity(),
            tfb.Identity(),
        ]

        # HMC transition kernel
        hmc_kernel = tfp.mcmc.HamiltonianMonteCarlo(
            target_log_prob_fn=self.target_log_prob_fn,
            step_size=step_size,
            num_leapfrog_steps=num_leapfrog_steps)
        hmc_kernel = tfp.mcmc.TransformedTransitionKernel(
            inner_kernel=hmc_kernel, bijector=constraining_bijectors)

        # Set default value for adaptation steps
        if num_adapt_steps is None:
            num_adapt_steps = int(0.8 * n_burnin)

        # Add step size adaptation
        hmc_kernel = tfp.mcmc.SimpleStepSizeAdaptation(
            inner_kernel=hmc_kernel, num_adaptation_steps=int(4000), target_accept_prob=0.8)

        # tracing function
        def trace_fn(_, pkr):
            return {
                'target_log_prob': pkr.inner_results.inner_results.accepted_results.target_log_prob,
                'diverging': (pkr.inner_results.inner_results.log_accept_ratio < -1000.),
                'is_accepted': pkr.inner_results.inner_results.is_accepted,
                'step_size': pkr.inner_results.inner_results.accepted_results.step_size,
            }

        # HMC sampling
        states, kernel_results, duration = self.sampling(num_results, n_burnin, hmc_kernel, self.params, trace_fn)

        # apply burnin
        states_burnin, sample_stats, acc_rate = self.get_chains_after_burnin(states, kernel_results, n_burnin)

        y_hat = self.get_y_hat(states_burnin, num_results, n_burnin)

        params = dict(zip(self.param_names, states_burnin))

        cell_types_nb = self.cell_types[:self.baseline_index] + self.cell_types[self.baseline_index + 1:]

        posterior = {var_name: [var] for var_name, var in params.items() if
                     "prediction" not in var_name}
        posterior_predictive = {"prediction": [params["prediction"]]}
        observed_data = {"y": self.y}
        dims = {"alpha": ["cell_type"],
                "b": ["covariate", "cell_type_nb"],
                "beta": ["covariate", "cell_type"],
                "concentration": ["sample", "cell_type"],
                "prediction": ["sample", "cell_type"]
                }
        coords = {"cell_type": self.cell_types,
                  "cell_type_nb": cell_types_nb,
                  "covariate": self.covariate_names,
                  "sample": range(self.y.shape[0])
                  }

        sampling_stats = {"chain_length": num_results, "n_burnin": n_burnin,
                        "acc_rate": acc_rate, "duration": duration, "y_hat": y_hat}

        model_specs = {"baseline": self.baseline_index, "formula": self.formula}


        return res.CAResultConverter(posterior=posterior,
                                     posterior_predictive=posterior_predictive,
                                     observed_data=observed_data,
                                     dims=dims,
                                     sample_stats=sample_stats,
                                     coords=coords).to_result_data(sampling_stats=sampling_stats,
                                                                   model_specs=model_specs)

    # Calculate predicted cell counts (for analysis purposes)
    def get_y_hat(self, states_burnin, num_results, n_burnin):
        """
        Calculate predicted cell counts (for analysis purposes) and add intermediate parameters to MCMC results

        Parameters
        ----------
        states_burnin -- list
            MCMC chain without burnin samples
        num_results -- int
            Chain length (with burnin)
        n_burnin -- int
            Number of burnin samples

        Returns
        -------
        y_mean
            predicted cell counts
        """
        chain_size_beta = [num_results - n_burnin, self.D, self.K]
        chain_size_y = [num_results - n_burnin, self.N, self.K]

        alphas = states_burnin[0]
        alphas_final = alphas.mean(axis=0)

        b = states_burnin[1]
        beta = np.zeros(chain_size_beta)
        for i in range(num_results - n_burnin):
            beta[i] = np.concatenate([b[i, :, :self.baseline_index],
                                      np.zeros(shape=[self.D, 1], dtype=np.float32),
                                      b[i, :, self.baseline_index:]], axis=1)

        betas_final = beta.mean(axis=0)

        conc_ = np.exp(np.einsum("jk, ...kl->...jl", self.x, beta)
                       + alphas.reshape((num_results - n_burnin, 1, self.K))).astype(np.float32)

        predictions_ = np.zeros(chain_size_y)
        for i in range(num_results - n_burnin):
            pred = tfd.DirichletMultinomial(self.n_total, conc_[i, :, :]).mean().numpy()
            predictions_[i, :, :] = pred

        states_burnin.append(beta)
        states_burnin.append(conc_)
        states_burnin.append(predictions_)

        concentration = np.exp(np.matmul(self.x, betas_final) + alphas_final).astype(np.float32)
        y_mean = concentration / np.sum(concentration, axis=1, keepdims=True) * self.n_total.numpy()[:, np.newaxis]
        return y_mean


class scdney_model:
    """
    wrapper for using the scdney package for R with scCODA data
    """

    def __init__(self, data):
        """
        Prepares R sampling

        Parameters
        ----------
        data -- scCODA data object
            scCODA data object
        ns -- list
            number of samples per condition

        Returns
        -------
        Creates .txt objects
        """

        # prepare list generation
        n, k = data.X.shape
        x_vec = data.X.flatten()
        cell_types = ["cell_" + x for x in data.var.index.tolist()]
        cell_types[0] = "cell_" + str(k)
        conditions = ["Cond_0", "Cond_1"]

        # get number of samples for both conditionas
        ns_0 = int(sum(data.obs["x_0"]))
        ns = [ns_0, n-ns_0]

        subjects = []
        for n in range(ns[0]):
            subjects.append("Cond_0_sub_" + str(n))
        for n in range(ns[1]):
            subjects.append("Cond_1_sub_" + str(n))

        # produce lists to use in scdney
        self.scdc_celltypes = []
        self.scdc_subject = []
        self.scdc_cond = []
        self.scdc_sample_cond = []

        for i in range(len(x_vec)):
            current_count = x_vec[i]
            current_type = cell_types[i % k]
            current_subject = subjects[i // k]
            current_condition = conditions[i // (k * ns[0])]

            self.scdc_sample_cond.append(current_condition)

            for j in range(int(current_count)):
                self.scdc_celltypes.append(current_type)
                self.scdc_subject.append(current_subject)
                self.scdc_cond.append(current_condition)

    def analyze(self, server=False):
        """
        Analyzes results from R script for SCDC from scdney packege.
        It is assumed that the effect on the first cell type is significant, all others are not.

        Returns
        -------
        Tuple:
            Tuple(raw summary from R, True positive...)
        """
        if server:
            os.environ["R_HOME"] = "/home/icb/johannes.ostner/anaconda3/lib/R"
            os.environ["PATH"] = r"/home/icb/johannes.ostner/anaconda3/lib/R/bin" + ";" + os.environ["PATH"]
        else:
            os.environ["R_HOME"] = "C:/Program Files/R/R-4.0.3"
            if "C:/Program Files/R/R-4.0.3/bin/x64" not in os.environ["PATH"]:
                os.environ["PATH"] = r"C:/Program Files/R/R-4.0.3/bin/x64" + ";" + os.environ["PATH"]

        import rpy2.robjects as rp
        from rpy2.robjects import numpy2ri, pandas2ri
        numpy2ri.activate()
        pandas2ri.activate()

        r_summary = rp.r(f"""
            library(scdney)
            library(tidyverse)
            library(broom.mixed)
            clust = scDC_noClustering({rp.vectors.StrVector(self.scdc_celltypes).r_repr()}, 
                                      {rp.vectors.StrVector(self.scdc_subject).r_repr()},
                                             calCI=TRUE,
                                             calCI_method=c("BCa"),
                                             nboot=100)

            glm = fitGLM(clust, {rp.vectors.StrVector(self.scdc_sample_cond).r_repr()}, pairwise=FALSE)
            sum = summary(glm$pool_res_random)
            print(sum)
            sum
            """)

        r_summary = pd.DataFrame(r_summary)

        p_values = r_summary.loc[r_summary["term"].str.contains("condCond_1"), "p.value"].values

        tp = np.sum(p_values[-1] < 0.05)
        fn = np.sum(p_values[-1] >= 0.05)
        tn = np.sum(p_values[:-1] >= 0.05)
        fp = np.sum(p_values[:-1] < 0.05)

        return r_summary, (tp, tn, fp, fn)


class FrequentistModel:
    """
    Superclass for making frequentist models in statsmodels from scCODA data (for model comparison)
    """

    def __init__(self, data):

        x = data.obs.to_numpy()
        y = data.X
        self.var = data.var

        self.x = x
        self.y = y
        self.n_total = np.sum(y, axis=1)

        self.p_val = {}

        # Get dimensions of data
        N = y.shape[0]

        # Check input data
        if N != x.shape[0]:
            raise ValueError("Wrong input dimensions X[{},:] != y[{},:]".format(y.shape[0], x.shape[0]))
        if N != len(self.n_total):
            raise ValueError("Wrong input dimensions X[{},:] != n_total[{}]".format(y.shape[0], len(self.n_total)))

    def eval_model(self, alpha=0.05, fdr_correct=True):
        """
        Evaluates array of p-values.
        It is assumed that the effect on the first cell type is significant, all others are not.

        Returns
        -------
        tp, tn, fp, fn : Tuple
            Number of True positive, ... effects
        """

        K = self.y.shape[1]
        ks = list(range(K))[1:]

        if fdr_correct:
            reject, pvals, _, _ = sm.stats.multitest.multipletests(self.p_val, alpha, method="fdr_bh")
            tp = sum([reject[0] == True])
            fn = sum([reject[0] == False])
            tn = sum([reject[k] == False for k in ks])
            fp = sum([reject[k] == True for k in ks])
        else:
            tp = sum([self.p_val[0] < alpha])
            fn = sum([self.p_val[0] >= alpha])
            tn = sum([self.p_val[k] >= alpha for k in ks])
            fp = sum([self.p_val[k] < alpha for k in ks])

        return tp, tn, fp, fn


class HaberModel(FrequentistModel):
    """
    Implements the Poisson regression model from Haber et al. into the scCODA framework
    (for model comparison purposes)
    """
    def fit_model(self):
        """
        Fits Poisson model

        Returns
        -------
        p_val : list
            p-values for differential abundance test of all cell types
        """

        p_val = []
        K = self.y.shape[1]

        if self.y.shape[0] == 2:
            p_val = [0 for x in range(K)]
        else:
            for k in range(K):
                data_ct = pd.DataFrame({"x": self.x[:, 0],
                                        "y": self.y[:, k]})

                model_ct = glm('y ~ x', data=data_ct, family=sm.genmod.families.Poisson(), offset=np.log(self.n_total)).fit()
                p_val.append(model_ct.pvalues[1])

        self.p_val = p_val


class CLRModel(FrequentistModel):
    """
    Implements a CLR transform and subsequent linear model on each cell type into the scCODA framework
    (for model comparison purposes)
    """
    def fit_model(self):
        """
        Fits CLR model

        Returns
        -------
        p_val : list
            p-values for differential abundance test of all cell types
        """

        p_val = []
        K = self.y.shape[1]
        D = self.x.shape[1]

        if self.y.shape[0] == 2:
            p_val = [0 for x in range(K)]
        else:
            # computes clr-transformed data matrix as a pandas DataFrame
            geom_mean = np.prod(self.y, axis=1, keepdims=True) ** (1 / K)
            y_clr = np.log(self.y / geom_mean)

            for k in range(K):
                data_ct = pd.DataFrame({"x": self.x[:, 0],
                                        "y": y_clr[:, k]})

                model_ct = glm('y ~ x', data=data_ct).fit()
                p_val.append(model_ct.pvalues[1])

        self.p_val = p_val


class TTest(FrequentistModel):
    """
        Implements a KS test one each cell type into the scCODA framework
        (for model comparison purposes)
    """

    def fit_model(self):
        """
        Fits KS test model

        Returns
        -------
        p_val : list
            p-values for differential abundance test of all cell types
        """

        p_val = []
        N, K = self.y.shape

        n_group = int(N/2)

        if self.y.shape[0] == 2:
            p_val = [0 for x in range(K)]
        else:
            for k in range(K):

                test = stats.ttest_ind(self.y[0:n_group, k], self.y[n_group:, k])
                p_val.append(test[1])

        self.p_val = p_val


class CLR_ttest(FrequentistModel):
    """
    Implements a CLR transform and subsequent linear model on each cell type into the scCODA framework
    (for model comparison purposes)
    """
    def fit_model(self):
        """
        Fits CLR model

        Returns
        -------
        p_val : list
            p-values for differential abundance test of all cell types
        """

        p_val = []
        N, K = self.y.shape
        D = self.x.shape[1]

        n_group = int(N/2)

        if self.y.shape[0] == 2:
            p_val = [0 for x in range(K)]
        else:
            # computes clr-transformed data matrix as a pandas DataFrame
            geom_mean = np.prod(self.y, axis=1, keepdims=True) ** (1 / K)
            y_clr = np.log(self.y / geom_mean)

            for k in range(K):
                test = stats.ttest_ind(y_clr[0:n_group, k], y_clr[n_group:, k])
                p_val.append(test[1])

        self.p_val = p_val


class ALDEx2Model(FrequentistModel):

    def fit_model(self, method="we.eBH", server=False, *args, **kwargs):

        if server:
            os.environ["R_HOME"] = "/home/icb/johannes.ostner/anaconda3/lib/R"
            os.environ["PATH"] = r"/home/icb/johannes.ostner/anaconda3/lib/R/bin" + ";" + os.environ["PATH"]
        else:
            os.environ["R_HOME"] = "C:/Program Files/R/R-4.0.3"
            os.environ["PATH"] = r"C:/Program Files/R/R-4.0.3/bin/x64" + ";" + os.environ["PATH"]

        K = self.y.shape[1]

        if self.y.shape[0] == 2:
            p_val = [0 for x in range(K)]
            self.result = None
        else:

            import rpy2.robjects as rp
            from rpy2.robjects import numpy2ri, pandas2ri
            numpy2ri.activate()
            pandas2ri.activate()
            import rpy2.robjects.packages as rpackages
            aldex2 = rpackages.importr("ALDEx2")

            cond = rp.vectors.FloatVector(self.x.astype("str").flatten().tolist())

            X_t = self.y.T
            nr, nc = X_t.shape
            X_r = rp.r.matrix(X_t, nrow=nr, ncol=nc)

            if "denom" in kwargs.keys():
                kwargs["denom"] = rp.vectors.FloatVector(kwargs["denom"])

            aldex_out = aldex2.aldex_clr(X_r, cond, *args, **kwargs)
            aldex_out = aldex2.aldex_ttest(aldex_out)
            aldex_out = pd.DataFrame(aldex_out)

            p_val = aldex_out.loc[:, method]

            self.result = aldex_out

        self.p_val = p_val


class ALRModel_ttest(FrequentistModel):
    """
    Implements a CLR transform and subsequent linear model on each cell type into the scCODA framework
    (for model comparison purposes)
    """
    def fit_model(self, reference_index):
        """
        Fits CLR model

        Returns
        -------
        p_val : list
            p-values for differential abundance test of all cell types
        """

        p_val = []
        N, K = self.y.shape
        D = self.x.shape[1]

        n_group = int(N/2)

        if self.y.shape[0] == 2:
            p_val = [0 for x in range(K)]
        else:
            # computes alr-transformed data matrix as a pandas DataFrame
            y_alr = np.log(self.y / self.y[:, reference_index][:, np.newaxis])

            for k in range(K):
                test = stats.ttest_ind(y_alr[0:n_group, k], y_alr[n_group:, k])
                p = test[1]
                if np.isnan(p):
                    p = 1
                p_val.append(p)

        self.p_val = p_val


class ALRModel_wilcoxon(FrequentistModel):
    """
    Implements a CLR transform and subsequent linear model on each cell type into the scCODA framework
    (for model comparison purposes)
    """
    def fit_model(self, reference_index):
        """
        Fits CLR model

        Returns
        -------
        p_val : list
            p-values for differential abundance test of all cell types
        """

        p_val = []
        N, K = self.y.shape
        D = self.x.shape[1]

        n_group = int(N/2)

        if self.y.shape[0] == 2:
            p_val = [0 for x in range(K)]
        else:
            # computes alr-transformed data matrix as a pandas DataFrame
            y_alr = np.log(self.y / self.y[:, reference_index][:, np.newaxis])

            for k in range(K):
                test = stats.ranksums(y_alr[0:n_group, k], y_alr[n_group:, k])
                p = test[1]
                if np.isnan(p):
                    p = 1
                p_val.append(p)

        self.p_val = p_val


class AncomModel():
    """
        Implements a KS test one each cell type into the scCODA framework
        (for model comparison purposes)
    """
    def __init__(self, data):
        x = data.obs
        y = pd.DataFrame(data.X, index=data.obs.index)
        self.x = x
        self.y = y
        self.n_total = np.sum(y, axis=1)

        self.ancom_out = []

        # Get dimensions of data
        N = y.shape[0]

        # Check input data
        if N != x.shape[0]:
            raise ValueError("Wrong input dimensions X[{},:] != y[{},:]".format(y.shape[0], x.shape[0]))
        if N != len(self.n_total):
            raise ValueError("Wrong input dimensions X[{},:] != n_total[{}]".format(y.shape[0], len(self.n_total)))

    def fit_model(self):
        """
        Fits KS test model

        Returns
        -------
        p_val : list
            p-values for differential abundance test of all cell types
        """

        K = self.y.shape[1]

        if self.y.shape[0] == 2:
            ancom_out = [False for x in range(K)]
        else:
            ancom_out = skbio.stats.composition.ancom(self.y, self.x["x_0"])

        self.ancom_out = ancom_out

    def eval_model(self):
        """
        Evaluates array of p-values.
        It is assumed that the effect on the first cell type is significant, all others are not.

        Returns
        -------
        tp, tn, fp, fn : Tuple
            Number of True positive, ... effects
        """
        K = self.y.shape[1]

        if self.y.shape[0] == 2:
            accept = [False for x in range(K)]
        else:
            accept = self.ancom_out[0]["Reject null hypothesis"].tolist()

        ks = list(range(K))[1:]

        tp = sum([accept[0]])
        fn = sum([accept[0] is False])
        tn = sum([accept[k] is False for k in ks])
        fp = sum([accept[k] for k in ks])

        return tp, tn, fp, fn


class DirichRegModel(FrequentistModel):

    def fit_model(self, server=False, *args, **kwargs):

        if server:
            os.environ["R_HOME"] = "/home/icb/johannes.ostner/anaconda3/lib/R"
            os.environ["PATH"] = r"/home/icb/johannes.ostner/anaconda3/lib/R/bin" + ";" + os.environ["PATH"]
        else:
            os.environ["R_HOME"] = "C:/Program Files/R/R-4.0.3"
            if "C:/Program Files/R/R-4.0.3/bin/x64" not in os.environ["PATH"]:
                os.environ["PATH"] = r"C:/Program Files/R/R-4.0.3/bin/x64" + ";" + os.environ["PATH"]

        K = self.y.shape[1]

        if self.y.shape[0] == 2:
            p_val = [0 for x in range(K)]
            self.result = None
        else:

            import rpy2.robjects as rp
            from rpy2.robjects import numpy2ri, pandas2ri
            numpy2ri.activate()
            pandas2ri.activate()

            p_val = rp.r(f"""
            library(DirichletReg)

            counts = {pandas2ri.py2rpy_pandasdataframe(pd.DataFrame(self.y, columns=self.var.index)).r_repr()}
            counts$counts = DR_data(counts)
            data = cbind(counts, {pandas2ri.py2rpy_pandasdataframe(pd.DataFrame(self.x, columns=["x_0"])).r_repr()})

            fit = DirichReg(counts ~ x_0, data)
            if(fit$optimization$convergence > 2L) {{
            pvals = matrix(c(0,0,0,0,0),nrow = 1)
            }} else {{
            u = summary(fit)
            pvals = u$coef.mat[grep('Intercept', rownames(u$coef.mat), invert=T), 4]
            v = names(pvals)
            pvals = matrix(pvals, ncol=length(u$varnames))
            rownames(pvals) = gsub('condition', '', v[1:nrow(pvals)])
            colnames(pvals) = u$varnames
            }}
            pvals
            """)
            p_val = p_val[0]

        self.p_val = p_val





"""
Toolbox for simulating compositional data from ScRNA-seq

This toolbox provides data generation and modelling solutions for compositional data with different specifications.
This data might e.g. come from scRNA-seq experiments.
For scenarios 1-4, we first generate composition parameters (b_true, w_true) and a covariance matrix (x) from some input specifications.
We then build a concentration vector for each sample (row of x) that sums up to 1. From there, we can calculate each row of the cell count matrix (y) via a multinomial distribution

:authors: Johannes Ostner
"""

import numpy as np
import anndata as ad
import pandas as pd
from scipy.special import softmax


def generate_normal_uncorrelated(N, D, K, n_total, noise_std_true=1):
    """
    Scenario 1: Normally distributed, independent covariates

    Parameters
    ----------
    N -- int
        Number of samples
    D -- int
        Number of covariates
    K -- int
        Number of cell types
    n_total -- list
        Number of individual cells per sample
    noise_std_true -- float
        noise level. 0: No noise

    Returns
    -------
    data
        Anndata object
    """

    # Generate random composition parameters
    b_true = np.random.normal(0, 1, size=K).astype(np.float32)  # bias (alpha)
    w_true = np.random.normal(0, 1, size=(D, K)).astype(np.float32)  # weights (beta)

    # Generate random covariate matrix
    x = np.random.normal(0, 1, size=(N, D)).astype(np.float32)
    noise = noise_std_true * np.random.randn(N, 1).astype(np.float32)

    # Generate y
    y = np.zeros([N, K], dtype=np.float32)
    for i in range(N):
        # Concentration should sum to 1 for each sample
        concentration = softmax(x[i, :].T@w_true + b_true + noise[i, :]).astype(np.float32)
        y[i, :] = np.random.multinomial(n_total[i], concentration).astype(np.float32)

    x_names = ["x_" + str(n) for n in range(x.shape[1])]
    x_df = pd.DataFrame(x, columns=x_names)

    data = ad.AnnData(X=y, obs=x_df, uns={"b_true": b_true, "w_true": w_true})

    return data


def generate_normal_correlated(N, D, K, n_total, noise_std_true, covariate_mean=None, covariate_var=None):
    """
    Scenario 2: Correlated covariates

    Parameters
    ----------
    N -- int
        Number of samples
    D -- int
        Number of covariates
    K -- int
        Number of cell types
    n_total -- list
        Number of individual cells per sample
    noise_std_true -- float
        noise level. 0: No noise
    covariate_mean -- numpy array [D]
        Mean of each covariate
    covariate_var -- numpy array [DxD]
        Covariance matrix for covariates

    Returns
    -------
    data
        Anndata object
    """

    if covariate_mean is None:
        covariate_mean = np.zeros(shape=D)

    # Generate randomized covariate covariance matrix if none is specified
    if covariate_var is None:
        # Covariates drawn from MvNormal(0, Cov), Cov_ij = p ^|i-j| , p=0.4
        # Tibshirani for correlated covariates: Tibshirani (1996)
        p = 0.4
        covariate_var = np.zeros((D, D))
        for i in range(D):
            for j in range(D):
                covariate_var[i, j] = p**np.abs(i-j)

    # Generate random composition parameters
    b_true = np.random.normal(0, 1, size=K).astype(np.float32)  # bias (alpha)
    w_true = np.random.normal(0, 1, size=(D, K)).astype(np.float32)  # weights (beta)

    # Generate random covariate matrix
    x = np.random.multivariate_normal(size=N, mean=covariate_mean, cov=covariate_var).astype(np.float32)
    noise = noise_std_true * np.random.randn(N, 1).astype(np.float32)

    # Generate y
    y = np.zeros([N, K], dtype=np.float32)
    for i in range(N):
        # Concentration should sum to 1 for each sample
        concentration = softmax(x[i, :].T @ w_true + b_true + noise[i, :]).astype(np.float32)
        y[i, :] = np.random.multinomial(n_total[i], concentration).astype(np.float32)

    x_names = ["x_" + str(n) for n in range(x.shape[1])]
    x_df = pd.DataFrame(x, columns=x_names)

    data = ad.AnnData(X=y, obs=x_df, uns={"b_true": b_true, "w_true": w_true})

    return data


def generate_normal_xy_correlated(N, D, K, n_total, noise_std_true=1,
                                  covariate_mean=None, covariate_var=None, sigma=None):
    """
    Scenario 3: Correlated cell types and covariates

    Parameters
    ----------
    N -- int
        Number of samples
    D -- int
        Number of covariates
    K -- int
        Number of cell types
    n_total -- list
        Number of individual cells per sample
    noise_std_true -- float
        noise level. 0: No noise
    covariate_mean -- numpy array [D]
        Mean of each covariate
    covariate_var -- numpy array [DxD]
        Covariance matrix for all covaraiates
    sigma -- numpy array [KxK]
        correlation matrix for cell types

    Returns
    -------
    data
        Anndata object
    """

    if covariate_mean is None:
        covariate_mean = np.zeros(shape=D)

    if sigma is None:
        sigma = np.identity(K)

    # Generate randomized covariate covariance matrix if none is specified
    if covariate_var is None:
        # Covaraits drawn from MvNormal(0, Cov) Cov_ij = p ^|i-j| , p=0.4
        # Tibshirani for correlated covariates: Tibshirani (1996)
        p = 0.4
        covariate_var = np.zeros((D, D))
        for i in range(D):
            for j in range(D):
                covariate_var[i, j] = p**np.abs(i-j)

    # Generate random composition parameters
    b_true = np.random.normal(0, 1, size=K).astype(np.float32)  # bias (alpha)
    w_true = np.random.normal(0, 1, size=(D, K)).astype(np.float32)  # weights (beta)

    # Generate random covariate matrix
    x = np.random.multivariate_normal(size=N, mean=covariate_mean, cov=covariate_var).astype(np.float32)
    noise = noise_std_true * np.random.randn(N, 1).astype(np.float32)

    # Generate y
    y = np.zeros([N, K], dtype=np.float32)
    for i in range(N):
        # Each row of y is now influenced by sigma
        alpha = np.random.multivariate_normal(mean=x[i, :].T@w_true + b_true, cov=sigma*noise[i, :]).astype(np.float32)
        concentration = softmax(alpha).astype(np.float32)
        y[i, :] = np.random.multinomial(n_total[i], concentration).astype(np.float32)

    x_names = ["x_" + str(n) for n in range(x.shape[1])]
    x_df = pd.DataFrame(x, columns=x_names)

    data = ad.AnnData(X=y, obs=x_df, uns={"b_true": b_true, "w_true": w_true})

    return data


def sparse_effect_matrix(D, K, n_d, n_k):
    """
    Generates a sparse effect matrix

    Parameters
    ----------
    D -- int
        Number of covariates
    K -- int
        Number of cell types
    n_d -- int
        Number of covariates that effect a cell type
    n_k -- int
        Number of cell types that are affected by any covariate

    Returns
    -------
    w_true
        Effect matrix
    """

    # Choose indices of affected cell types and covariates randomly
    d_eff = np.random.choice(range(D), size=n_d, replace=False)
    k_eff = np.random.choice(range(K), size=n_k, replace=False)

    # Possible entries of w_true
    w_choice = [0.3, 0.5, 1]
    
    w_true = np.zeros((D, K))
    # Fill in w_true
    for i in d_eff:
        for j in k_eff:
            c = np.random.choice(3, 1)
            w_true[i, j] = w_choice[c]
            
    return w_true
    

def generate_sparse_xy_correlated(N, D, K, n_total, noise_std_true=1,
                                  covariate_mean=None, covariate_var=None,
                                  sigma=None,
                                  b_true=None, w_true=None):
    """
    Scenario 4: Sparse true parameters

    Parameters
    ----------
    N -- int
        Number of samples
    D -- int
        Number of covariates
    K -- int
        Number of cell types
    n_total -- list
        Number of individual cells per sample
    noise_std_true -- float
        noise level. 0: No noise
    covariate_mean -- numpy array [D]
        Mean of each covariate
    covariate_var -- numpy array [DxD]
        Covariance matrix for all covaraiates
    sigma -- numpy array [KxK]
        correlation matrix for cell types
    b_true -- numpy array [K]
        bias coefficients
    w_true -- numpy array [DxK]
        Effect matrix

    Returns
    -------
    data
        Anndata object
    """

    if covariate_mean is None:
        covariate_mean = np.zeros(shape=D)

    if sigma is None:
        sigma = np.identity(K)

    # Generate randomized covariate covariance matrix if none is specified
    if covariate_var is None:
        # Covaraits drawn from MvNormal(0, Cov) Cov_ij = p ^|i-j| , p=0.4
        # Tibshirani for correlated covariates: Tibshirani (1996)
        p = 0.4
        covariate_var = np.zeros((D, D))
        for i in range(D):
            for j in range(D):
                covariate_var[i, j] = p ** np.abs(i - j)

    # Uniform intercepts if none are specifed
    if b_true is None:
        b_true = np.random.uniform(-3,3, size=K).astype(np.float32)  # bias (alpha)

    # Randomly select covariates that should correlate if none are specified
    if w_true is None:
        n_d = np.random.choice(range(D), size=1)
        n_k = np.random.choice(range(K), size=1)
        w_true = sparse_effect_matrix(D, K, n_d, n_k)

    # Generate random covariate matrix
    x = np.random.multivariate_normal(size=N, mean=covariate_mean, cov=covariate_var).astype(np.float32)
    noise = noise_std_true * np.random.randn(N, 1).astype(np.float32)

    # Generate y
    y = np.zeros([N, K], dtype=np.float32)
    for i in range(N):
        # Each row of y is now influenced by sigma
        alpha = np.random.multivariate_normal(mean=x[i, :].T @ w_true + b_true, cov=sigma * noise[i, :]).astype(
            np.float32)
        concentration = softmax(alpha).astype(np.float32)
        y[i, :] = np.random.multinomial(n_total[i], concentration).astype(np.float32)

    x_names = ["x_" + str(n) for n in range(x.shape[1])]
    x_df = pd.DataFrame(x, columns=x_names)

    data = ad.AnnData(X=y, obs=x_df, uns={"b_true": b_true, "w_true": w_true})

    return data


def generate_case_control(cases=1, K=5, n_total=1000, n_samples=[5,5], noise_std_true=0,
                          sigma=None, b_true=None, w_true=None):
    """
    Generates compositional data with binary covariates

    Parameters
    ----------
    cases -- int
        number of covariates
    K -- int
        Number of cell types
    n_total -- int
        number of cells per sample
    n_samples -- list
        Number of samples per case combination as array[2**cases]
    noise_std_true -- float
        noise level. 0: No noise - Not in use atm!!!
    sigma -- numpy array [KxK]
        correlation matrix for cell types
    b_true -- numpy array [K]
        bias coefficients
    w_true -- numpy array [DxK]
        Effect matrix

    Returns
    -------
    Anndata object
    """
    D = cases**2

    # Uniform intercepts if none are specifed
    if b_true is None:
        b_true = np.random.uniform(-3, 3, size=K).astype(np.float32)  # bias (alpha)

    # Randomly select covariates that should correlate if none are specified
    if w_true is None:
        n_d = np.random.choice(range(D), size=1)
        n_k = np.random.choice(range(K), size=1)
        w_true = sparse_effect_matrix(D, K, n_d, n_k)

    # Sigma is identity if not specified else
    if sigma is None:
        sigma = np.identity(K) * 0.05

    # noise = noise_std_true * np.random.randn(N, 1).astype(np.float32)

    # Initialize x, y
    x = np.zeros((sum(n_samples), cases))
    y = np.zeros((sum(n_samples), K)) 
    c = 0

    # Binary representation of x as list of fixed length
    def binary(x, length):
        return [int(x_n) for x_n in bin(x)[2:].zfill(length)]

    # For all combinations of cases
    for i in range(2**cases):
        # For each sample with this combination
        for j in range(n_samples[i]):
            # row of x is binary representation
            x[c+j] = binary(i, cases)

            # Generate y
            alpha = np.random.multivariate_normal(mean=x[c+j, :].T @ w_true + b_true, cov=sigma).astype(
                np.float32)

            concentration = softmax(alpha).astype(np.float32)
            z = np.random.multinomial(n_total, concentration)
            y[c+j] = z
        c = c+n_samples[i]

    x = x.astype(np.float32)
    y = y.astype(np.float32)

    x_names = ["x_" + str(n) for n in range(x.shape[1])]
    x_df = pd.DataFrame(x, columns=x_names)

    data = ad.AnnData(X=y, obs=x_df, uns={"b_true": b_true, "w_true": w_true})

    return data


def b_w_from_abs_change(counts_before=np.array([200, 200, 200, 200, 200]), abs_change=50, n_total=1000):
    """
    Calculates intercepts and slopes from a starting count and an absolute change for the first cell type

    Parameters
    ----------
    counts_before -- numpy array
        cell counts for control samples
    abs_change -- int
        change of first cell type in terms of cell counts
    n_total -- int
        number of cells per sample. This stays constant over all samples!!!

    Returns
    -------
    intercepts -- numpy array
        intercept parameters
    slopes -- numpy array
        slope parameters
    """

    K = counts_before.shape[0]

    # calculate intercepts for control samples
    b = np.log(counts_before / n_total)

    # count vector after applying the effect.
    # sum(counts_after) = n_total;
    # counts_after[0] = counts_before[0] + abs_change
    count_0_after = counts_before[0] + abs_change
    count_other_after = (n_total - count_0_after) / (K - 1)
    counts_after = np.repeat(count_other_after, K)
    counts_after[0] = count_0_after

    # Get parameter vector with effect
    b_after = np.log(counts_after / n_total)

    # w is the difference of b before and after
    w = b_after - b
    # Transform w such that only first entry is nonzero
    w = w - w[K - 1]

    return b, w


def counts_from_first(b_0=200, n_total=1000, K=5):
    b = np.repeat((n_total-b_0)/(K-1), K)
    b[0] = b_0
    return b



"""
Unit tests for scCODA
"""

import unittest
import numpy as np
import scanpy as sc
import tensorflow as tf
import pandas as pd
import os
import sys
sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('..'))

from sccoda.util import cell_composition_data as dat
from sccoda.util import comp_ana as mod
from sccoda.util import data_generation as gen

pd.set_option('display.max_columns', 500)
pd.set_option('display.max_rows', 500)

#%%


class TestDataGeneration(unittest.TestCase):
    """
    Testing whether the data generation functions from data_generation work as intended
    Returns
    -------
    boolean -- all tests were passed or not
    """

    def setUp(self):

        self.N = 3
        self.D = 1
        self.K = 2
        self.n_total = [1000] * self.N
        self.noise_std_true = 1
        self.covariate_mean = None
        self.covariate_var = None
        self.sigma = None
        self.b_true = None
        self.w_true = None

    def test_case_1(self):

        np.random.seed(1234)

        test_1 = True
        data_sc_1 = gen.generate_normal_uncorrelated(self.N, self.D, self.K, self.n_total, self.noise_std_true)
        if any(np.abs(data_sc_1.obs["x_0"] - [-0.720589, 0.887163, 0.859588]) > 1e-5):
            print("scenario1.obs is not correct!")
            test_1 = False
        if not np.array_equal(data_sc_1.X, np.array([[591., 409.], [959., 41.], [965.,  35.]])):
            print("scenario1.X is not correct!")
            test_1 = False

        self.assertTrue(test_1)

    def test_case_2(self):
        np.random.seed(1234)

        test_2 = True
        data_sc_2 = gen.generate_normal_correlated(self.N, self.D, self.K, self.n_total, self.noise_std_true,
                                                   self.covariate_mean, self.covariate_var)
        if any(np.abs(data_sc_2.obs["x_0"] - [-0.720589, 0.887163, 0.859588]) > 1e-5):
            print("scenario2.obs is not correct!")
            test_2 = False
        if not np.array_equal(data_sc_2.X, np.array([[591., 409.], [959., 41.], [965., 35.]])):
            print("scenario2.X is not correct!")
            test_2 = False

        self.assertTrue(test_2)

    def test_case_3(self):
        np.random.seed(1234)

        test_3 = True
        data_sc_3 = gen.generate_normal_xy_correlated(self.N, self.D, self.K, self.n_total, self.noise_std_true,
                                                      self.covariate_mean, self.covariate_var, self.sigma)
        if any(np.abs(data_sc_3.obs["x_0"] - [-0.720589, 0.887163, 0.859588]) > 1e-5):
            print("scenario3.obs is not correct!")
            test_3 = False
        if not np.array_equal(data_sc_3.X, np.array([[570., 430.], [957., 43.], [923., 77.]])):
            print("scenario3.X is not correct!")
            test_3 = False

        self.assertTrue(test_3)

    def test_case_4(self):
        np.random.seed(1234)

        test_4 = True
        data_sc_4 = gen.generate_sparse_xy_correlated(self.N, self.D, self.K, self.n_total, self.noise_std_true,
                                                      self.covariate_mean, self.covariate_var, self.sigma,
                                                      self.b_true, self.w_true)
        if any(np.abs(data_sc_4.obs["x_0"] - [0.662509, 0.675246, -0.940298]) > 1e-5):
            print("scenario4.obs is not correct!")
            test_4 = False
        if not np.array_equal(data_sc_4.X, np.array([[12., 988.], [13., 987.], [26., 974.]])):
            print("scenario4.X is not correct!")
            test_4 = False

        self.assertTrue(test_4)

    def test_case_control_gen(self):
        """
        Tests data generation for case/control scenarios
        Returns
        -------
        boolean -- all tests were passed or not
        """
        np.random.seed(1234)

        cases = 1
        K = 2
        n_total = 1000
        n_samples = [2, 2]
        noise_std_true = 0
        sigma = None
        b_true = None
        w_true = None

        data = gen.generate_case_control(cases, K, n_total, n_samples, noise_std_true, sigma, b_true, w_true)

        test = True
        if any(np.abs(data.obs["x_0"] - [0, 0, 1, 1]) > 1e-5):
            print("obs is not correct!")
            test = False
        if not np.array_equal(data.X, np.array([[74., 926.], [58., 942.], [32., 968.], [53., 947.]])):
            print("X is not correct!")
            test = False
        if not np.array_equal(data.uns["b_true"], np.array([-1.8508832,  0.7326526], dtype=np.float32)) & \
           np.array_equal(data.uns["w_true"], np.array([[0., 0.]])):
            print("uns is not correct!")
            test = False

        self.assertTrue(test)

    def test_change_functions(self):
        """
        Tests gen.b_w_from_abs_change and gen.counts_from_first
        Returns
        -------
        boolean -- all tests were passed or not
        """
        np.random.seed(1234)
        correct = True

        counts_before = np.array([600, 400])
        abs_change = 100
        n_total = 1000
        K = 2
        b_0 = 600

        b, w = gen.b_w_from_abs_change(counts_before, abs_change, n_total)

        if any(np.abs(b - [-0.51082562, -0.91629073]) > 1e-5):
            print("gen.b_w_from_abs_change: b not correct!")
            correct = False

        if any(np.abs(w - [0.44183275, 0.]) > 1e-5):
            print("gen.b_w_from_abs_change: b not correct!")
            correct = False

        b_2 = gen.counts_from_first(b_0, n_total, K)
        if not np.array_equal(b_2, [600., 400.]):
            print("gen.counts_from_first not correct!")
            correct = False

        self.assertTrue(correct)


class TestDataImport(unittest.TestCase):

    def test_from_pandas(self):
        # Get Haber Salmonella data
        data_raw = pd.read_csv("./data/haber_counts.csv")

        salm_indices = [0, 1, 2, 3, 8, 9]
        salm_df = data_raw.iloc[salm_indices, :]

        data_salm = dat.from_pandas(salm_df, covariate_columns=["Mouse"])
        data_salm.obs["Condition"] = data_salm.obs["Mouse"].str.replace(r"_[0-9]", "")

        # Only check size of x, obs
        x_shape = (data_salm.X.shape == (6, 8))
        obs_shape = (data_salm.obs.shape == (6, 2))

        self.assertTrue(x_shape & obs_shape)

    def test_from_scanpy(self):
        # Get scanpy example data, add covariates, read in three times
        adata_ref = sc.datasets.pbmc3k_processed()
        adata_ref.uns["cov"] = {"x1": 0, "x2": 1}

        data = dat.from_scanpy_list([adata_ref, adata_ref, adata_ref],
                                    cell_type_identifier="louvain",
                                    covariate_key="cov")

        # Only check size of x, obs
        x_shape = (data.X.shape == (3, 8))
        obs_shape = (data.obs.shape == (3, 2))
        var_names = (data.var.index.tolist() == ['CD4 T cells', 'CD14+ Monocytes', 'B cells', 'CD8 T cells',
                                                 'NK cells', 'FCGR3A+ Monocytes', 'Dendritic cells', 'Megakaryocytes'])

        self.assertTrue(x_shape & obs_shape & var_names)


class TestModels(unittest.TestCase):

    def setUp(self):

        # Get Haber count data
        data_raw = pd.read_csv("./data/haber_counts.csv")

        salm_indices = [0, 1, 2, 3, 8, 9]
        salm_df = data_raw.iloc[salm_indices, :]

        data_salm = dat.from_pandas(salm_df, covariate_columns=["Mouse"])
        data_salm.obs["Condition"] = data_salm.obs["Mouse"].str.replace(r"_[0-9]", "")
        self.data = data_salm

    def test_no_baseline(self):
        np.random.seed(1234)
        tf.random.set_seed(5678)

        model_salm = mod.CompositionalAnalysis(self.data, formula="Condition", baseline_index=None)

        # Run MCMC
        sim_results = model_salm.sample_hmc(num_results=20000, n_burnin=5000)
        alpha_df, beta_df = sim_results.summary_prepare()

        # Mean cell counts for both groups
        alphas_true = np.round(np.mean(self.data.X[:4], 0), 0)
        betas_true = np.round(np.mean(self.data.X[4:], 0), 0)

        # Mean cell counts for simulated data
        final_alphas = np.round(alpha_df.loc[:, "expected_sample"].tolist(), 0)
        final_betas = np.round(beta_df.loc[:, "expected_sample"].tolist(), 0)

        # Check if model approximately predicts ground truth
        differing_alphas = any(np.abs(alphas_true - final_alphas) > 30)
        differing_betas = any(np.abs(betas_true - final_betas) > 30)

        self.assertTrue((not differing_alphas) & (not differing_betas))

    def test_baseline(self):
        np.random.seed(1234)
        tf.random.set_seed(5678)

        model_salm = mod.CompositionalAnalysis(self.data, formula="Condition", baseline_index=5)

        # Run MCMC
        sim_results = model_salm.sample_hmc(num_results=20000, n_burnin=5000)
        alpha_df, beta_df = sim_results.summary_prepare()

        # Mean cell counts for both groups
        alphas_true = np.round(np.mean(self.data.X[:4], 0), 0)
        betas_true = np.round(np.mean(self.data.X[4:], 0), 0)

        # Mean cell counts for simulated data
        final_alphas = np.round(alpha_df.loc[:, "expected_sample"].tolist(), 0)
        final_betas = np.round(beta_df.loc[:, "expected_sample"].tolist(), 0)

        # Check if model approximately predicts ground truth
        differing_alphas = any(np.abs(alphas_true - final_alphas) > 30)
        differing_betas = any(np.abs(betas_true - final_betas) > 30)

        self.assertTrue((not differing_alphas) & (not differing_betas))


#%%

if __name__ == '__main__':
    unittest.main()
